#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os

from zenml.version import __version__


def handle_bool_env_var(var, default=False):
    """Converts normal env var to boolean"""
    var = os.getenv(var, None)
    for i in ['1', 'y', 'yes', 'True', 'true']:
        if i == var:
            return True
    return default


# Global constants
APP_NAME = 'zenml'
CONFIG_VERSION = '1'

# Environment variables
ENV_ZENML_DEBUG = 'ZENML_DEBUG'
ENV_ZENML_LOGGING_VERBOSITY = 'ZENML_LOGGING_VERBOSITY'
ENV_ABSL_LOGGING_VERBOSITY = 'ZENML_ABSL_LOGGING_VERBOSITY'

# Logging variables
IS_DEBUG_ENV = handle_bool_env_var(ENV_ZENML_DEBUG, default=False)
ZENML_LOGGING_VERBOSITY = os.getenv(ENV_ZENML_LOGGING_VERBOSITY, 3)
ABSL_LOGGING_VERBOSITY = os.getenv(ENV_ABSL_LOGGING_VERBOSITY, -100)

# Base images for zenml
ZENML_REGISTRY = 'eu.gcr.io/maiot-zenml'
ZENML_BASE_IMAGE_NAME = f'{ZENML_REGISTRY}/zenml:base-{__version__}'
ZENML_TRAINER_IMAGE_NAME = f'{ZENML_REGISTRY}/zenml:cuda-{__version__}'
ZENML_DATAFLOW_IMAGE_NAME = f'{ZENML_REGISTRY}/zenml:dataflow-{__version__}'

# Evaluation utils constants
COMPARISON_NOTEBOOK = 'comparison_notebook.ipynb'
EVALUATION_NOTEBOOK = 'evaluation_notebook.ipynb'

# Pipeline related constants
PREPROCESSING_FN = 'zenml.components.transform.transform_module.preprocessing_fn'
TRAINER_FN = 'zenml.components.trainer.trainer_module.run_fn'

# GCP Orchestration
GCP_ENTRYPOINT = 'zenml.backends.orchestrator.entrypoint'
AWS_ENTRYPOINT = 'zenml.backends.orchestrator.entrypoint'
K8S_ENTRYPOINT = 'zenml.backends.orchestrator.entrypoint'



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from typing import List

import tensorflow as tf

from zenml.steps.trainer import TFFeedForwardTrainer
from zenml.steps.trainer import utils
from zenml.utils import naming_utils


class ImageTensorflowTrainer(TFFeedForwardTrainer):
    """Generic Trainer for Images"""

    def __init__(self,
                 batch_size: int = 8,
                 lr: float = 0.001,
                 epochs: int = 1,
                 loss: str = 'mse',
                 metrics: List[str] = None,
                 hidden_activation: str = 'relu',
                 last_activation: str = 'sigmoid',
                 **kwargs):
        """
        Basic feedforward neural network constructor.

        Args:
            batch_size: Input data batch size.
            lr: Learning rate of the optimizer.
            epochs: Number of training epochs (whole passes through the data).
            metrics: List of metrics to log in training.
            hidden_layers: List of sizes to use in consecutive hidden layers.
             Length determines the number of hidden layers in the network.
            hidden_activation: Name of the activation function to use in the
             hidden layers.
            last_activation: Name of the final activation function creating
             the class probability distribution.

            **kwargs: Additional keyword arguments.
        """

        self.batch_size = batch_size
        self.lr = lr
        self.epochs = epochs
        self.loss = loss
        self.metrics = metrics or []
        self.hidden_activation = hidden_activation
        self.last_activation = last_activation

        super(ImageTensorflowTrainer, self).__init__(
            batch_size=batch_size,
            lr=lr,
            epochs=epochs,
            loss=loss,
            metrics=metrics,
            hidden_activation=hidden_activation,
            last_activation=last_activation,
            **kwargs
        )

    def run_fn(self):
        train_split_patterns = [self.input_patterns[split] for split in
                                self.split_mapping[utils.TRAIN_SPLITS]]
        train_dataset = self.input_fn(train_split_patterns)

        eval_split_patterns = [self.input_patterns[split] for split in
                               self.split_mapping[utils.EVAL_SPLITS]]
        eval_dataset = self.input_fn(eval_split_patterns)

        model = self.model_fn(train_dataset=train_dataset,
                              eval_dataset=eval_dataset)

        signatures = self.get_signatures(model)

        model.save(
            self.serving_model_dir, save_format='tf', signatures=signatures)

    def model_fn(self,
                 train_dataset: tf.data.Dataset,
                 eval_dataset: tf.data.Dataset):
        """
        Simple function that defines a CNN for MNIST.

        Args:
            train_dataset: tf.data.Dataset containing the training data.
            eval_dataset: tf.data.Dataset containing the evaluation data.

        Returns:
            model: A trained feedforward neural network model.
        """
        features = list(train_dataset.element_spec[0].keys())
        labels = list(train_dataset.element_spec[1].keys())

        features = [k for k in features if
                    naming_utils.check_if_transformed_feature(k)]
        full_shape = train_dataset.element_spec[0][features[0]].shape

        input = tf.keras.layers.Input(shape=full_shape[1:], name=features[0])
        x = tf.keras.layers.Conv2D(16, kernel_size=(3, 3), activation="relu")(
            input)
        x = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(x)
        x = tf.keras.layers.Flatten()(x)
        x = tf.keras.layers.Dense(1, name=labels[0])(x)

        model = tf.keras.Model(inputs=input, outputs=x)

        model.compile(loss=self.loss,
                      optimizer=tf.keras.optimizers.Adam(lr=self.lr),
                      metrics=self.metrics)

        model.summary()

        tensorboard_callback = tf.keras.callbacks.TensorBoard(
            log_dir=self.log_dir)

        model.fit(train_dataset,
                  epochs=self.epochs,
                  validation_data=eval_dataset,
                  callbacks=[tensorboard_callback])

        return model



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import tensorflow as tf

import tensorflow_transform as tft


def _impute(tensor, replacement):
    """
    Args:
        tensor:
        replacement:
    """
    if isinstance(tensor, tf.sparse.SparseTensor):
        sparse = tf.sparse.SparseTensor(tensor.indices,
                                        tensor.values,
                                        [tensor.dense_shape[0], 1])
        return tf.sparse.to_dense(sp_input=sparse, default_value=replacement)
    else:
        return tensor


def custom_f(custom_value):
    """
    Args:
        custom_value:
    """

    def apply(x):
        x = _impute(x, custom_value)
        return x

    return apply


def min_f():
    def apply(x):
        m = tft.min(x)
        x = _impute(x, m)
        return x

    return apply


def max_f():
    def apply(x):
        m = tft.max(x)
        x = _impute(x, m)
        return x

    return apply


def mean_f():
    def apply(x):
        m = tft.mean(x)

        x = tf.cast(x, dtype=tf.float32)
        x = _impute(x, m)
        return x

    return apply



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, List, Text, Any

import tensorflow as tf

from zenml.standards.standard_keys import MethodKeys, DefaultKeys
from zenml.steps.preprocesser import BasePreprocesserStep
from zenml.steps.preprocesser.standard_preprocesser.methods.standard_methods \
    import NonSeqFillingMethods, TransformMethods
from zenml.utils import naming_utils
from zenml.utils.preprocessing_utils import parse_methods, DEFAULT_DICT


def infer_schema(schema):
    """
    Function to infer a schema from input data.
    Args:
        schema: Input dict mapping features to tf.Tensors.

    Returns:
        schema_dict: Dict mapping features to their respective data types.
    """
    schema_dict = dict()
    for feature, descriptions in schema.items():
        dtype = descriptions.dtype
        # These are ugly hacks for hard checking whether its broadly int, float
        #  bool or string.
        if dtype == float or 'float' in str(dtype):
            schema_dict[feature] = DefaultKeys.FLOAT
        elif dtype == int or 'int' in str(dtype):
            schema_dict[feature] = DefaultKeys.INTEGER
        elif dtype == bool or 'bool' in str(dtype):
            schema_dict[feature] = DefaultKeys.BOOLEAN
        elif dtype == str or 'string' in str(dtype):
            schema_dict[feature] = DefaultKeys.STRING
        else:
            raise AssertionError('Cannot infer schema dtype!')
    return schema_dict


class StandardPreprocesser(BasePreprocesserStep):
    """
    Standard Preprocessor step. This step can be used to apply a variety of
    standard preprocessing techniques from the field of Machine Learning,
    which are predefined in ZenML, to the data.
    """

    def __init__(self,
                 features: List[Text] = None,
                 labels: List[Text] = None,
                 overwrite: Dict[Text, Any] = None,
                 **unused_kwargs):
        """
        Standard preprocessing step for generic preprocessing. Preprocessing
        steps are inferred on a feature-by-feature basis from its data type,
        and different standard preprocessing methods are applied based on the
        data type of the feature.

        The overwrite argument can also be used to overwrite the data
        type-specific standard preprocessing functions and apply other
        functions instead, given that they are registered in the default
        preprocessing method dictionary defined at the top of this file.

        An entry in the `overwrite` dict could look like this: ::

            {feature_name:
                {"transform": [{"method": "scale_to_z_score",
                                           "parameters": {}}],
                 "filling": [{"method": "max", "parameters": {}}]
            }
        
        Args:
            features: List of data features to be preprocessed.
            labels: List of features in the data that are to be predicted.
            overwrite: Dict of dicts, mapping features to a list of
             custom preprocessing and filling methods to be used.
            **unused_kwargs: Additional unused keyword arguments. Their usage
             might change in the future.
        """
        super(StandardPreprocesser, self).__init__(features=features,
                                                   labels=labels,
                                                   overwrite=overwrite,
                                                   **unused_kwargs)

        # List of features and labels
        self.features = features or []
        self.labels = labels or []

        self.t_dict = parse_methods(overwrite or {},
                                    'transform',
                                    TransformMethods)

        self.t_d_dict = parse_methods(DEFAULT_DICT,
                                      'transform',
                                      TransformMethods)

        self.f_dict = parse_methods(overwrite or {},
                                    'filling',
                                    NonSeqFillingMethods)

        self.f_d_dict = parse_methods(DEFAULT_DICT,
                                      'filling',
                                      NonSeqFillingMethods)

    def preprocessing_fn(self, inputs: Dict):
        """
        Standard preprocessing function.
        Args:
            inputs: Dict mapping features to their respective data.

        Returns:
            output: Dict mapping transformed features to their respective
             transformed data.

        """
        schema = infer_schema(inputs)

        output = {}
        for key, value in inputs.items():
            if key not in self.f_dict:
                self.f_dict[key] = self.f_d_dict[schema[key]]
            value = self.apply_filling(value, self.f_dict[key])

            if key in self.features or key in self.labels:
                if key not in self.t_dict:
                    self.t_dict[key] = self.t_d_dict[schema[key]]
                result = self.apply_transform(key, value, self.t_dict[key])
                result = tf.cast(result, dtype=tf.float32)

                if key in self.features:
                    output[naming_utils.transformed_feature_name(key)] = result
                if key in self.labels:
                    output[naming_utils.transformed_label_name(key)] = result

            output[key] = value

        return output

    @staticmethod
    def apply_transform(key, data, transform_list):
        """
        Apply a list of transformations to input data.
        Args:
            key: Key argument specific to vocabulary computation.
            data: Data to be input into the transform functions.
            transform_list: List of transforms to apply to the data.

        Returns:
            data: Transformed data after each of the transforms in
             transform_list have been applied.

        """
        for transform in transform_list:
            method_name = transform[MethodKeys.METHOD]
            method_params = transform[MethodKeys.PARAMETERS]

            if method_name == 'compute_and_apply_vocabulary':
                method_params.update({'vocab_filename': key})

            data = TransformMethods.get_method(method_name)(data,
                                                            **method_params)
        return data

    @staticmethod
    def apply_filling(data, filling_list):
        """
        Apply a list of fillings to input data.
        Args:
            data: Data to be input into the transform functions.
            filling_list: List of fillings to apply to the data. As of now,
             only the first filling in the list will be applied.

        Returns:
            data: Imputed data after the first filling in filling_list
            has been applied.

        """
        method_name = filling_list[0][MethodKeys.METHOD]
        method_params = filling_list[0][MethodKeys.PARAMETERS]
        return NonSeqFillingMethods.get_method(method_name)(**method_params)(
            data)

    def get_preprocessing_fn(self):
        return self.preprocessing_fn



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import json
from pathlib import Path
from typing import Text, Dict

import yaml

from zenml.utils import path_utils


def write_yaml(file_path: Text, contents: Dict):
    """
    Write contents as YAML format to file_path.

    Args:
        file_path (str): Path to YAML file.
        contents (dict): Contents of YAML file as dict.
    """
    if not path_utils.is_remote(file_path):
        dir_ = str(Path(file_path).parent)
        if not path_utils.is_dir(dir_):
            # If it is a local path and it doesnt exist, raise Exception.
            raise Exception(f'Directory {dir_} does not exist.')
    path_utils.write_file_contents(file_path, yaml.dump(contents))


def read_yaml(file_path: Text):
    """
    Read YAML on file path and returns contents as dict.

    Args:
        file_path (str): Path to YAML file.
    """
    if path_utils.file_exists(file_path):
        with open(file_path, 'r') as f:
            return yaml.load(f.read(), Loader=yaml.FullLoader)
    else:
        raise Exception(f'{file_path} does not exist.')


def is_yaml(file_path: Text):
    """
    Returns True if file_path is YAML, else False

    Args:
        file_path (str): Path to YAML file.
    """
    if file_path.endswith('yaml') or file_path.endswith('yml'):
        return True
    return False


def write_json(file_path: Text, contents: Dict):
    """
    Write contents as JSON format to file_path.

    Args:
        file_path (str): Path to JSON file.
        contents (dict): Contents of JSON file as dict.
    """
    if not path_utils.is_remote(file_path):
        dir_ = str(Path(file_path).parent)
        if not path_utils.is_dir(dir_):
            # If it is a local path and it doesnt exist, raise Exception.
            raise Exception(f'Directory {dir_} does not exist.')
    path_utils.write_file_contents(file_path, json.dumps(contents))


def read_json(file_path: Text):
    """
    Read JSON on file path and returns contents as dict.

    Args:
        file_path (str): Path to JSON file.
    """
    if path_utils.file_exists(file_path):
        with open(file_path, 'r') as f:
            return json.loads(f.read())
    else:
        raise Exception(f'{file_path} does not exist.')



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import json
import os
from pathlib import Path

import pytest

import zenml
from zenml.components.data_gen.constants import FILE_EXT, FILE_NAME, \
    BINARY_DATA, IMAGE, LABEL
from zenml.datasources.image_datasource import ImageDatasource
from zenml.datasources.image_datasource import get_matching_label, \
    add_label_and_metadata, SplitByFileName

ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")

json_data = {"1f22663e72.jpg": {"label": 0,
                                "metadata": {"height": 256,
                                             "width": 256,
                                             "num_channels": 3}}}


def test_image_local_commit(repo):
    data = Path(__file__).resolve() \
               .parent.parent.parent / 'tests' / 'test_data' / 'images'
    ds = ImageDatasource(
        name='image_ds_local',
        base_path=str(data.resolve()),
    )
    assert not ds.commits

    _id = ds.commit()

    # check that this commit made
    assert _id == ds.get_latest_commit()


def test_image_gcp_commit(repo):
    ds = ImageDatasource(
        name='image_ds_gcp',
        base_path='gs://zenml_quickstart/cycle_gan_mini',
    )
    assert not ds.commits

    _id = ds.commit()

    # check that this commit made
    assert _id == ds.get_latest_commit()


def test_get_matching_label():
    filename = "1f22663e72.jpg"
    label_data = json.dumps(json_data)

    label, metadata = get_matching_label(label_data=label_data,
                                         img_filename=filename)

    expected_label = json_data[filename]["label"]
    expected_metadata = json_data[filename]["metadata"]

    assert label == expected_label
    assert metadata == expected_metadata

    # simulate an image filename not found inside the dict
    json_data2 = {}
    label_data = json.dumps(json_data2)

    with pytest.raises(AssertionError):
        _, _ = get_matching_label(label_data=label_data, img_filename=filename)


def test_file_split():
    # example image
    example1 = {FILE_NAME: "abc", FILE_EXT: ".jpg"}

    # example label file, found by the file extension .txt
    example2 = {FILE_NAME: "def", FILE_EXT: ".txt"}

    # pathological image file, taken as label file by the function because
    # the word "label" is present in the file name
    example3 = {FILE_NAME: "xyz_label", FILE_EXT: ".jpg"}

    # png image, to be detected as image
    example4 = {FILE_NAME: "data", FILE_EXT: ".png"}

    splits = [SplitByFileName(ex, 2) for ex in [example1, example2,
                                                example3, example4]]

    assert splits == [0, 1, 1, 0]


def test_add_label_and_metadata():
    filename = "1f22663e72.jpg"
    label_data = {BINARY_DATA: json.dumps(json_data)}

    example_img = {FILE_NAME: filename, BINARY_DATA: b"12345",
                   FILE_EXT: '.jpg'}

    updated_img = add_label_and_metadata(image_dict=example_img,
                                         label_dict=label_data)

    assert BINARY_DATA not in updated_img
    assert IMAGE in updated_img
    assert LABEL in updated_img
    assert all(md_key in updated_img for md_key in
               json_data[filename]["metadata"])



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
import random
from pathlib import Path

import pytest

import zenml
from zenml import exceptions
from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.datasources import BaseDatasource, ImageDatasource
from zenml.enums import PipelineStatusTypes, GDPComponent
from zenml.pipelines import BasePipeline, TrainingPipeline
from zenml.standards import standard_keys as keys
from zenml.steps import BaseStep
from zenml.utils import path_utils

# Nicholas a way to get to the root
ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")


def test_executed(repo):
    name = "my_pipeline"
    p = BasePipeline(name=name)
    assert not p.is_executed_in_metadata_store

    random_run_pipeline = random.choice(repo.get_pipelines())

    assert random_run_pipeline.is_executed_in_metadata_store


def test_naming():
    name = "my_pipeline"
    p = BasePipeline(name=name)
    # file_name = p.file_name
    pipeline_name = p.pipeline_name

    assert p.get_name_from_pipeline_name(pipeline_name) == name


def test_get_status(repo):
    name = "my_pipeline"
    p = BasePipeline(name=name)

    # passing this test means that if the pipeline is not in the metadata
    # store, it is (semantically) not started yet
    assert p.get_status() == PipelineStatusTypes.NotStarted.name

    run_pipeline = random.choice(repo.get_pipelines())

    assert run_pipeline.get_status() == PipelineStatusTypes.Succeeded.name


def test_register_pipeline(repo, delete_config):
    name = "my_pipeline"
    p: BasePipeline = BasePipeline(name=name)

    p._check_registered()

    random_run_pipeline = random.choice(repo.get_pipelines())
    with pytest.raises(exceptions.AlreadyExistsException):
        random_run_pipeline._check_registered()

    p.register_pipeline({"name": name})

    delete_config(p.file_name)


def test_add_datasource():
    name = "my_pipeline"
    p: BasePipeline = BasePipeline(name=name)

    p.add_datasource(BaseDatasource(name="my_datasource"))

    assert isinstance(p.datasource, BaseDatasource)

    # assert not p.steps_dict[keys.TrainingSteps.DATA]


def test_pipeline_copy(repo):
    random_run_pipeline = random.choice(repo.get_pipelines())

    new_name = "my_second_pipeline"

    new_pipeline = random_run_pipeline.copy(new_name=new_name)

    assert not new_pipeline._immutable
    assert isinstance(new_pipeline, random_run_pipeline.__class__)


def test_get_pipeline_config():
    # test just the pipeline config block added to the steps_config
    name = "my_pipeline"
    p: BasePipeline = BasePipeline(name=name)

    config = p.get_pipeline_config()

    p_name = p.pipeline_name

    p_args = config[keys.PipelineKeys.ARGS]

    assert p_args[keys.PipelineDetailKeys.NAME] == p_name
    # assert p_args[keys.PipelineDetailKeys.TYPE] == "base"
    assert p_args[keys.PipelineDetailKeys.ENABLE_CACHE] is True
    assert config[keys.PipelineKeys.DATASOURCE] == {}
    assert config[keys.PipelineKeys.SOURCE].split("@")[0] == \
           "zenml.pipelines.base_pipeline.BasePipeline"
    # TODO: Expand this to more pipelines


def test_get_steps_config():
    # TODO: Expand this to more steps
    name = "my_pipeline"
    p: BasePipeline = BasePipeline(name=name)

    kwargs = {"number": 1, "description": "abcdefg"}
    step = BaseStep(**kwargs)

    p.steps_dict["test"] = step

    cfg = p.get_steps_config()

    steps_cfg = cfg[keys.PipelineKeys.STEPS]

    # avoid missing args / type inconsistencies
    assert steps_cfg["test"] == step.to_config()


def test_get_artifacts_uri_by_component(repo):
    test_component_name = GDPComponent.SplitGen.name

    ps = repo.get_pipelines_by_type([TrainingPipeline.PIPELINE_TYPE])

    p: BasePipeline = ps[0]

    uri_list = p.get_artifacts_uri_by_component(test_component_name)

    # assert it is not empty
    assert uri_list
    # assert artifact was written
    uri = uri_list[0]
    written_artifacts = path_utils.list_dir(uri)
    assert written_artifacts
    # TODO: Ugly TFRecord validation
    assert all((("tfrecord" in name and
                 os.path.splitext(name)[-1] == ".gz") for name in f)
               for _, _, f in os.walk(uri))


def test_to_from_config(equal_pipelines):
    p1: BasePipeline = BasePipeline(name="my_pipeline")

    ds = ImageDatasource(name="my_datasource")

    p1.add_datasource(ds)

    p2 = BasePipeline.from_config(p1.to_config())

    assert equal_pipelines(p1, p2, loaded=True)


def test_load_config(repo, equal_pipelines):
    p1 = random.choice(repo.get_pipelines())

    pipeline_config = p1.load_config()

    p2 = BasePipeline.from_config(pipeline_config)

    assert equal_pipelines(p1, p2, loaded=True)


def test_run_config():
    p = BasePipeline(name="my_pipeline")

    class MockBackend(OrchestratorBaseBackend):
        def run(self, config):
            return {"message": "Run triggered!"}

    # run config without a specified backend
    p.run_config(p.to_config())

    p.backend = "123"

    with pytest.raises(Exception):
        # not a backend subclass error
        p.run_config(p.to_config())

    p.backend = MockBackend()

    assert not p.run_config(p.to_config())


def test_run_base(delete_config):
    # Test of pipeline.run(), without artifact / metadata store change
    p = BasePipeline(name="my_pipeline")

    class MockBackend(OrchestratorBaseBackend):
        def run(self, config):
            return {"message": "Run triggered!"}

    backend = MockBackend()

    p.run(backend=backend)

    assert p._immutable

    delete_config(p.file_name)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import click
from tabulate import tabulate

from zenml.cli.cli import cli
from zenml.cli.utils import pass_repo
from zenml.repo import Repository


@cli.group()
def step():
    """Steps group"""
    pass


@step.command("list")
@pass_repo
def list_steps(repo: Repository):
    step_versions = repo.get_step_versions()
    name_version_data = []
    headers = ["step_name", "step_version"]
    for name, version_set in step_versions.items():
        names = [name] * len(version_set)
        versions = list(version_set)
        name_version_data.extend(list(zip(names, versions)))

    click.echo(tabulate(name_version_data, headers=headers))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the DataFlow Orchestrator Backend"""

import os
from typing import Any, Text, Dict

from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.backends.orchestrator.beam.zenml_beam_orchestrator import \
    ZenMLBeamlDagRunner
from zenml.pipelines.utils import parse_yaml_beam_args


class OrchestratorBeamBackend(OrchestratorBaseBackend):
    """Uses Apache Beam as a Pipeline orchestrator."""

    def __init__(
            self,
            worker_machine_type: Text = 'e2-medium',
            num_workers: int = 4,
            max_num_workers: int = 10,
            disk_size_gb: int = 100,
            autoscaling_algorithm: Text = 'THROUGHPUT_BASED',
            **kwargs):
        self.worker_machine_type = worker_machine_type
        self.num_workers = num_workers
        self.max_num_workers = max_num_workers
        self.disk_size_gb = disk_size_gb
        self.autoscaling_algorithm = autoscaling_algorithm
        super().__init__(
            worker_machine_type=worker_machine_type,
            num_workers=num_workers,
            max_num_workers=max_num_workers,
            disk_size_gb=disk_size_gb,
            autoscaling_algorithm=autoscaling_algorithm,
            **kwargs
        )
        raise NotImplementedError('Its coming soon!')

    def run(self, config: Dict[Text, Any]):
        """
        This run function essentially calls an underlying TFX orchestrator run.
        However it is meant as a higher level abstraction with some
        opinionated decisions taken.

        Args:
            config: a ZenML config dict
        """
        beam_orchestrator_args = parse_yaml_beam_args({
            'worker_machine_type': self.worker_machine_type,
            'num_workers': self.num_workers,
            'max_num_workers': self.max_num_workers,
            'disk_size_gb': self.disk_size_gb,
            'autoscaling_algorithm': self.autoscaling_algorithm,
            'setup_file': os.path.join(os.getcwd(), 'setup.py'),
            # 'job_name': 'zenml-' + run_name,
            # 'temp_location': pipeline_temp,
            # 'staging_location': staging_location,
            # 'extra_package': gz_path,
            # 'requirements_file': req_path,
        })
        tfx_pipeline = self.get_tfx_pipeline(config)
        ZenMLBeamlDagRunner(beam_orchestrator_args).run(tfx_pipeline)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the base Training Backend"""

from zenml.components.trainer.executor import ZenMLTrainerExecutor
from tfx.dsl.components.base import executor_spec

from zenml.backends import BaseBackend


class TrainingBaseBackend(BaseBackend):
    """
    Base class for all base Training ZenML backends.

    Every ZenML pipeline runs in backends.

    A training backend can be used to efficiently train a machine learning
    model on large amounts of data. Since most common machine learning models
    leverage mainly linear algebra operations under the hood, they can
    potentially benefit a lot from dedicated training hardware like Graphics
    Processing Units (GPUs) or application-specific integrated circuits
    (ASICs).
    """
    BACKEND_TYPE = 'training'

    def get_executor_spec(self):
        """Return a TFX Executor spec for the Trainer Component."""
        return executor_spec.ExecutorClassSpec(ZenMLTrainerExecutor)

    def get_custom_config(self):
        """Return a dict to be passed as a custom_config to the Trainer."""
        return {}



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text

from tfx.extensions.google_cloud_ai_platform.pusher import \
    executor as ai_platform_pusher_executor

from zenml.steps.deployer import BaseDeployerStep
from zenml.logger import get_logger

logger = get_logger(__name__)


class GCAIPDeployer(BaseDeployerStep):
    """
    Step class for deploying models on Google Cloud AI Platform.
    """

    def __init__(self,
                 model_name: Text,
                 project_id: Text,
                 **kwargs):
        """
        Google Cloud AI Platform Deployer Step constructor.

        Use this step to push your trained model to Google Cloud AI Platform,
        where it gets assigned an endpoint that you can query for evaluations
        and predictions.

        Args:
            model_name: Name of the model.
            project_id: Name of the Google Cloud project which the model
             should be pushed to.
             deployment_path: Path where deployment is pushed.
            **kwargs: Additional keyword arguments.
        """
        self.project_id = project_id
        self.model_name = model_name
        logger.warning('When using GCAIPDeployer, please ensure that the '
                       'Artifact Store is a Google Cloud Bucket.')
        super(GCAIPDeployer, self).__init__(project_id=project_id,
                                            model_name=model_name,
                                            **kwargs)

    def get_config(self):
        ai_platform_serving_args = {
            'model_name': self.model_name,
            'project_id': self.project_id
        }
        return {'ai_platform_serving_args': ai_platform_serving_args}

    def get_executor(self):
        return ai_platform_pusher_executor.Executor



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ratio-based categorical split."""

from typing import Text, Dict, List, Any, Union

from zenml.steps.split import BaseSplitStep
from zenml.steps.split import constants
from zenml.steps.split.utils import get_categorical_value
from zenml.steps.split.utils import partition_cat_list

CategoricalValue = Union[Text, int]


def lint_split_map(split_map: Dict[Text, float]):
    """Small utility to lint the split_map"""
    if len(split_map) <= 1:
        raise AssertionError('Please specify more than 1 split name in the '
                             'split_map!')


class CategoricalRatioSplit(BaseSplitStep):
    """
    Categorical ratio split. Use this to split data based on a list of values
    of interest in a single categorical column. A categorical column is
    defined here as a column with finitely many values of type `integer` or
    `string`. In contrast to the categorical domain split, here categorical
    values are assigned to different splits by the corresponding percentages,
    defined inside a split ratio object.
    """

    def __init__(
            self,
            categorical_column: Text,
            categories: List[CategoricalValue],
            split_ratio: Dict[Text, float],
            unknown_category_policy: Text = constants.SKIP,
            statistics=None,
            schema=None,
    ):
        """
        Categorical domain split constructor.

        Use this class to split your data based on values in
        a single categorical column. A categorical column is defined here as a
        column with finitely many values of type `integer` or `string`.

        Example usage:

        # Split on a categorical attribute called "color", with a defined list
        of categories of interest

        # half of the categories go entirely into the train set,
          the other half into the eval set. Other colors, e.g. "purple",
           are discarded due to the "skip" flag.

        >>> split = CategoricalRatioSplit(
        ... categorical_column="color",
        ... categories = ["red", "green", "blue", "yellow"],
        ... split_ratio = {"train": 0.5,
        ...                "eval": 0.5},
        ... unknown_category_policy="skip")

        Supply the unknown_category_policy flag to set the unknown category
        handling policy. There are two main options:

        Setting unknown_category_policy to any key in the split map indicates
        that any missing categories should be put into that particular split.
        For example, supplying ``unknown_category_policy="train"`` indicates
        that all missing categories should go into the training dataset, while
        ``unknown_category_policy="eval"`` indicates that all missing
        categories should go into the evaluation dataset.

        Setting ``unknown_category_policy="skip"`` indicates that data points
        with unknown categorical values (i.e., values not present in the
        categorical value list) should be taken out of the data set.

        Args:
            statistics: Parsed statistics from a preceding StatisticsGen.
            schema: Parsed schema from a preceding SchemaGen.
            categorical_column: Name of the categorical column used for
             splitting.
            categories: List of categorical values found in the categorical
             column on which to split.
            split_ratio: A dict mapping { split_name: ratio of categories
             in split }.
            unknown_category_policy: String, indicates how to handle categories
             in the data that are not present in the supplied category list.
        """
        self.categorical_column = categorical_column

        split_map = partition_cat_list(categories, split_ratio)
        lint_split_map(split_map)
        self.split_map = split_map

        if unknown_category_policy in self.split_map:
            self.unknown_category_policy = unknown_category_policy
        else:
            self.unknown_category_policy = constants.SKIP

        super().__init__(statistics=statistics,
                         schema=schema,
                         categorical_column=categorical_column,
                         split_ratio=split_ratio,
                         categories=categories,
                         unknown_category_policy=unknown_category_policy)

    def partition_fn(self,
                     element: Any,
                     num_partitions: int) -> int:

        """
        Function for a categorical split on data to be used in a beam.Partition.
        Args:
            element: Data point, given as a tf.train.Example.
            num_partitions: Number of splits, unused here.

        Returns:
            An integer n, where 0 ≤ n ≤ num_partitions - 1.

        """

        category_value = get_categorical_value(element,
                                               cat_col=self.categorical_column)

        # The following code produces a dict: { split_name: unique_integer }
        enumerated_splits = {name: i for i, name in
                             enumerate(self.split_map.keys())}

        if self.unknown_category_policy not in enumerated_splits:
            enumerated_splits.update(
                {self.unknown_category_policy: num_partitions - 1})

        for split_name, category_value_list in self.split_map.items():
            # if the value is in the list, then just return
            if category_value in category_value_list:
                # return the index of that split
                return enumerated_splits[split_name]

        # This is the default behavior for category_values that dont belong to
        #  any split in the split_map.
        return enumerated_splits[self.unknown_category_policy]

    def get_split_names(self) -> List[Text]:
        split_names = list(self.split_map.keys())
        if self.unknown_category_policy in self.split_map:
            return split_names
        else:
            return split_names + [constants.SKIP]



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base Class for all ZenML datasources"""

import os
from abc import abstractmethod
from pathlib import Path
from typing import Text, Dict, Optional, Callable
from uuid import uuid4

import tensorflow as tf

from zenml.enums import GDPComponent
from zenml.exceptions import AlreadyExistsException
from zenml.exceptions import EmptyDatasourceException
from zenml.logger import get_logger
from zenml.metadata import ZenMLMetadataStore
from zenml.repo import Repository, ArtifactStore
from zenml.standards import standard_keys as keys
from zenml.utils import path_utils
from zenml.utils import source_utils
from zenml.utils.analytics_utils import CREATE_DATASOURCE
from zenml.utils.analytics_utils import track
from zenml.utils.post_training.post_training_utils import \
    view_schema, get_feature_spec_from_schema, \
    convert_raw_dataset_to_pandas, view_statistics
from zenml.utils.print_utils import to_pretty_string, PrintStyles

logger = get_logger(__name__)


class BaseDatasource:
    """Base class for all ZenML datasources.
    Every ZenML datasource should override this class.
    """

    def __init__(
            self,
            name: Text,
            _id: Text = None,
            backend=None,
            metadata_store: Optional[ZenMLMetadataStore] = None,
            artifact_store: Optional[ArtifactStore] = None,
            commits: Optional[Dict] = None,
            *args,
            **kwargs):
        """
        Construct the datasource.

        Args:
            name (str): name of datasource
            schema (dict): schema of datasource
            _id: unique ID (for internal use)
        """
        if _id:
            # Its loaded from config
            self._id = _id
            logger.debug(f'Datasource {name} loaded.')
        else:
            # If none, then this is assumed to be 'new'. Check dupes.
            all_names = Repository.get_instance().get_datasource_names()
            if any(d == name for d in all_names):
                raise AlreadyExistsException(
                    name=name,
                    resource_type='datasource')
            self._id = str(uuid4())
            track(event=CREATE_DATASOURCE)
            logger.info(f'Datasource {name} created.')

        # Metadata store
        if metadata_store:
            self.metadata_store: ZenMLMetadataStore = metadata_store
        else:
            # use default
            self.metadata_store: ZenMLMetadataStore = \
                Repository.get_instance().get_default_metadata_store()

        # Default to local
        if backend is None:
            from zenml.backends.orchestrator import OrchestratorBaseBackend
            self.backend = OrchestratorBaseBackend()
        else:
            self.backend = backend

        # Artifact store
        if artifact_store:
            self.artifact_store = artifact_store
        else:
            # use default
            self.artifact_store = \
                Repository.get_instance().get_default_artifact_store()

        if commits is None:
            self.commits = {}
        else:
            self.commits = commits

        self.name = name
        self._immutable = False
        self._source = source_utils.resolve_class(self.__class__)
        self._source_args = kwargs

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @property
    def is_empty(self):
        if self.commits:
            return False
        return True

    @property
    def n_datapoints(self):
        """Gets total number of datapoints in datasource"""
        pipeline = self._get_one_pipeline()
        data_files = self._get_data_file_paths(pipeline)
        return sum(1 for _ in tf.data.TFRecordDataset(data_files,
                                                      compression_type='GZIP'))

    @abstractmethod
    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        pass

    def commit(self):
        from zenml.pipelines.data_pipeline import DataPipeline
        data_pipeline = DataPipeline(
            enable_cache=False,
            backend=self.backend,
            metadata_store=self.metadata_store,
            artifact_store=self.artifact_store,
            datasource=self
        )
        data_pipeline.run()
        commit_id = data_pipeline.pipeline_name.split('_')[2]
        self.commits[commit_id] = data_pipeline.pipeline_name.split('_')[1]
        return commit_id

    def _assert_commit_id(self, commit_id: Text):
        """Asserts commit_id is in self.commits"""
        if commit_id not in self.commits:
            raise AssertionError(
                f'There is no such commit_id as {commit_id} in the '
                f'datasource {self.name}')

    @classmethod
    def from_config(cls, config: Dict):
        """
        Convert from Data Step config to ZenML Datasource object.
        Data step is also populated and configuration set to parameters set
        in the config file.
        Args:
            config: a DataStep config in dict-form (probably loaded from YAML).
        """
        if keys.DatasourceKeys.SOURCE not in config[
            keys.PipelineKeys.DATASOURCE]:
            return None  # can be empty

        # this is the data step config block
        source = config[keys.PipelineKeys.DATASOURCE][
            keys.DatasourceKeys.SOURCE]
        datasource_class = source_utils.load_source_path_class(source)
        datasource_name = config[keys.PipelineKeys.DATASOURCE][
            keys.DatasourceKeys.NAME]
        _id = config[keys.PipelineKeys.DATASOURCE][keys.DatasourceKeys.ID]
        args = config[keys.PipelineKeys.DATASOURCE][keys.DatasourceKeys.ARGS]

        # start with artifact store
        artifact_store = ArtifactStore(config[keys.PipelineKeys.DATASOURCE][
                                           keys.DatasourceKeys.ARTIFACT_STORE])

        # metadata store
        metadata_store: ZenMLMetadataStore = ZenMLMetadataStore.from_config(
            config=config[keys.PipelineKeys.DATASOURCE][
                keys.DatasourceKeys.METADATA_STORE]
        )

        # backend
        from zenml.backends.orchestrator import OrchestratorBaseBackend
        backend = OrchestratorBaseBackend.from_config(
            config=config[keys.PipelineKeys.DATASOURCE][
                keys.DatasourceKeys.BACKEND]
        )

        # resolve commits
        data_pipeline_names = \
            metadata_store.get_data_pipeline_names_from_datasource_name(
            datasource_name)
        # ugly hack to recompile the commit times
        commits = {}
        if data_pipeline_names:
            commits = {x.split('_')[2]: x.split('_')[1] for x in
                       data_pipeline_names}

        obj = datasource_class(
            name=datasource_name, _id=_id, commits=commits, backend=backend,
            metadata_store=metadata_store, artifact_store=artifact_store,
            **args)
        obj._immutable = True
        return obj

    def to_config(self):
        """Converts datasource to ZenML config block."""
        return {
            keys.DatasourceKeys.NAME: self.name,
            keys.DatasourceKeys.SOURCE: self._source,
            keys.DatasourceKeys.ARGS: self._source_args,
            keys.DatasourceKeys.ID: self._id,
            keys.DatasourceKeys.METADATA_STORE:
                self.metadata_store.to_config(),
            keys.DatasourceKeys.ARTIFACT_STORE: self.artifact_store.path,
            keys.DatasourceKeys.BACKEND: self.backend.to_config()
        }

    def get_latest_commit(self):
        a = [k for k, v in
             sorted(self.commits.items(), key=lambda item: item[1])]
        if a:
            return a[-1]

    def get_first_commit(self):
        a = [k for k, v in
             sorted(self.commits.items(), key=lambda item: item[1])]
        if a:
            return a[0]

    def get_data_pipeline_from_commit(self, commit_id: Text):
        from zenml.pipelines.data_pipeline import DataPipeline

        self._assert_commit_id(commit_id)
        repo: Repository = Repository.get_instance()
        name = DataPipeline.get_name_from_pipeline_name(
            DataPipeline.PIPELINE_TYPE + '_' + self.commits[
                commit_id] + '_' + commit_id)
        return repo.get_pipeline_by_name(name)

    def _get_one_pipeline(self):
        """Gets representative pipeline from all pipelines associated."""
        if self.commits:
            return self.get_data_pipeline_from_commit(
                list(self.commits.keys())[0])
        raise EmptyDatasourceException

    def _get_data_file_paths(self, pipeline):
        """
        Gets path where data is stored as list of file paths.
        Args:
            pipeline: a pipeline with this datasource embedded
        """
        if pipeline.datasource._id != self._id:
            raise AssertionError('This pipeline does not belong to this '
                                 'datasource.')
        # Take any pipeline and get the datagen
        data_uri = os.path.join(pipeline.get_artifacts_uri_by_component(
            GDPComponent.DataGen.name
        )[0], 'examples')
        data_files = path_utils.list_dir(data_uri)
        return data_files

    def sample_data(self, sample_size: int = 100000):
        """
        Sampels data from datasource as a pandas DataFrame.
        Args:
            sample_size: # of rows to sample.
        """
        pipeline = self._get_one_pipeline()
        data_files = self._get_data_file_paths(pipeline)

        schema_uri = pipeline.get_artifacts_uri_by_component(
            GDPComponent.DataSchema.name)[0]
        spec = get_feature_spec_from_schema(schema_uri)

        dataset = tf.data.TFRecordDataset(data_files, compression_type='GZIP')
        return convert_raw_dataset_to_pandas(dataset, spec, sample_size)

    def view_schema(self, commit_id: Text = None):
        """
        View schema of data flowing in pipeline.

        Args:
            commit_id: used to specify which commit's schema to use, if None
            uses latest
        """
        if commit_id is None:
            commit_id = self.get_latest_commit()
        self._assert_commit_id(commit_id)

        pipeline = self.get_data_pipeline_from_commit(commit_id)
        uri = pipeline.get_artifacts_uri_by_component(
            GDPComponent.DataSchema.name)[0]
        view_schema(uri)

    def view_statistics(self, commit_id: Text = None, port: int = None,
                        magic: bool = False):
        """
        View statistics of data flowing in pipeline.

        Args:
            port (int): Port at which to launch the statistics facet.
            commit_id: used to specify which commit's schema to use, if None
            uses latest
            magic (bool): Whether to display within a jupyter notebook or not
        """
        if commit_id is None:
            commit_id = self.get_latest_commit()
        self._assert_commit_id(commit_id)
        pipeline = self.get_data_pipeline_from_commit(commit_id)
        uri = pipeline.get_artifacts_uri_by_component(
            GDPComponent.DataStatistics.name)[0]
        view_statistics(uri, port=port, magic=magic)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Factory to register metadata_wrapper classes to metadata_wrappers"""

from zenml.metadata import MockMetadataStore
from zenml.metadata import MySQLMetadataStore
from zenml.metadata import SQLiteMetadataStore


class MetadataWrapperFactory:
    """Definition of MetadataWrapperFactory to track all metadata_wrappers.

    All metadata_wrappers (including custom metadata_wrappers) are to be
    registered here.
    """

    def __init__(self):
        self.metadata_wrappers = {}

    def get_metadata_wrappers(self):
        return self.metadata_wrappers

    def get_single_metadata_wrapper(self, key):
        return self.metadata_wrappers[key]

    def register_metadata_wrapper(self, key, metadata_wrapper_):
        self.metadata_wrappers[key] = metadata_wrapper_


# Register the injections into the factory
wrapper_factory = MetadataWrapperFactory()
wrapper_factory.register_metadata_wrapper(SQLiteMetadataStore.STORE_TYPE,
                                          SQLiteMetadataStore)
wrapper_factory.register_metadata_wrapper(MySQLMetadataStore.STORE_TYPE,
                                          MySQLMetadataStore)
wrapper_factory.register_metadata_wrapper(MockMetadataStore.STORE_TYPE,
                                          MockMetadataStore)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Image Datasource definition"""

import json
import os
from typing import Callable, Dict, Text, Any

import apache_beam as beam
from apache_beam.io import fileio

from zenml.components.data_gen.constants import BINARY_DATA, FILE_NAME, \
    FILE_EXT, METADATA, LABEL, IMAGE
from zenml.datasources import BaseDatasource
from zenml.utils.beam_utils import WriteToTFRecord


def read_file_content(file: beam.io.fileio.ReadableFile):
    """
    Read contents from a file handle in binary and return it along with some
    file metadata as a dict.

    Args:
        file (beam.io.fileio.ReadableFile): Beam ReadableFile object,
        corresponds to an image file read from disk.

    Returns:
        data_dict: Dict with binary data and file metadata.
    """
    file_path = file.metadata.path
    base = os.path.basename(file_path)

    data_dict = {BINARY_DATA: file.read(),
                 FILE_NAME: base,
                 FILE_EXT: os.path.splitext(base)[1]}

    return data_dict


def add_label_and_metadata(image_dict: Dict[Text, Any],
                           label_dict: Dict[Text, Any]):
    """
    Add label and metadata information to an image.

    Args:
        image_dict: Dict with image features.
        label_dict (Text): JSON-readable string with label information.

    Returns:
        image_dict: Updated image feature dict with label and metadata
         information.
    """
    filename = image_dict[FILE_NAME]

    label_data = label_dict[BINARY_DATA]
    label, metadata = get_matching_label(label_data, filename)

    image_dict.update(metadata)
    image_dict[LABEL] = label

    # pop binary image data and reinsert as "image"
    # for a more intuitive image feature name
    image_raw = image_dict.pop(BINARY_DATA)
    image_dict[IMAGE] = image_raw

    return image_dict


def get_matching_label(label_data: Text, img_filename: Text):
    """
    Get a label matching an image file name from a JSON-readable label file.

    Args:
        label_data (Text): Label string, needs to be JSON-readable.
        img_filename (Text): File name of the image.

    Returns:
        label: Label key of the image.
        metadata: Dict, additional metadata information.
    """

    # This can potentially go bad if the JSON is too big
    data_dict = json.loads(label_data)

    if img_filename not in data_dict:
        raise AssertionError(
            f'You need to provide a label for {img_filename} in your '
            f'labels.json!')
    img_data = data_dict[img_filename]

    if "label" not in img_data:
        raise ValueError("Error while parsing label file: No label "
                         "key was found. Make sure to specify your "
                         "image label under a \"label\" key.")

    label = img_data.pop(LABEL)

    try:
        metadata = img_data.pop(METADATA)
    except KeyError:
        metadata = {}

    # JSON parsing puts empty metadata dict as the empty string,
    # so in that case coerce to dict just to be type safe
    if not metadata:
        metadata = {}

    return label, metadata


@beam.typehints.with_input_types(Dict[Text, Any], int)
@beam.typehints.with_output_types(int)
def SplitByFileName(element: Dict[Text, Any],
                    num_partitions: int) -> int:
    """
    Helper function to identify the label file in a beam.Partition applied
    to the PCollection of input files.

    Args:
        element: Dict with image features.
        num_partitions (int): Number of partitions, unused.
    """
    filename = element[FILE_NAME]
    file_ext = element[FILE_EXT]

    if "label" in filename:
        return 1

    if file_ext.lower() in [".txt", ".json", ".csv"]:
        return 1

    return 0


class ImageDatasource(BaseDatasource):
    """ZenML Image datasource definition.

    Use this for image training pipelines.
    """

    def __init__(
            self,
            name: Text = None,
            base_path: Text = None,
            schema: Dict = None,
            **kwargs):
        """
        Create a Image datasource.

        Args:
            name (str): Name of datasource
            base_path (str): Path to folder of images
            schema (str): Optional schema for data to conform to.
        """
        self.base_path = base_path
        self.schema = schema
        super().__init__(name, base_path=base_path, schema=schema, **kwargs)

    def process(self, output_path: Text, make_beam_pipeline: Callable = None):
        wildcard_qualifier = "*"

        # ingest all the files from the base path by supplying the wildcard
        file_pattern = os.path.join(self.base_path, wildcard_qualifier)

        allowed_ext = [".jpg", ".json", ".png", ".txt", ".jpeg"]

        with make_beam_pipeline() as p:
            images, label_file = (
                    p
                    | fileio.MatchFiles(file_pattern)
                    | fileio.ReadMatches()
                    | beam.Map(read_file_content)
                    | "FilterOutFiles" >> beam.Filter(lambda x:
                                                      x[FILE_EXT] in
                                                      allowed_ext)
                    | "SplitLabelFile" >> beam.Partition(
                SplitByFileName,
                2))

            # label_file is actually a dict
            label_dict = beam.pvalue.AsSingleton(label_file)
            _ = (
                    images
                    | "AddLabelAndMetadata" >> beam.Map(add_label_and_metadata,
                                                        label_dict)
                    | WriteToTFRecord(self.schema, output_path))



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from datetime import datetime

import numpy as np
import pandas as pd
import pytz

from zenml.standards.standard_keys import MethodKeys, DefaultKeys
from zenml.steps.sequencer.standard_sequencer.methods import methods_resampling
from zenml.steps.sequencer.standard_sequencer.methods.standard_methods \
    import ResamplingMethods


def convert_datetime_to_secs(dt):
    delta = dt - datetime(year=1970, month=1, day=1, tzinfo=pytz.utc)
    return int(delta.total_seconds())


def array_to_value(value):
    if isinstance(value, np.ndarray) and value.size == 1:
        return value[0]
    return None


def get_resample_output_dtype(resampling_config, schema):
    output_dtype_map = {}
    for feature, description in resampling_config.items():
        method = description[0][MethodKeys.METHOD]
        params = description[0][MethodKeys.PARAMETERS]

        output_dtype_map[feature] = get_output_dtype(
            input_dtype=schema[feature],
            method=ResamplingMethods.get_method(method),
            params=params)

    return output_dtype_map


def get_output_dtype(input_dtype, method, params):
    if method == methods_resampling.resample_thresholding:
        if type(params['set_value']) is str:
            return str
        else:
            return float
    if input_dtype == 'string':
        if method == methods_resampling.resample_mode:
            return str
    return float


def extract_sequences(session, seq_length, freq, shift):
    sequence_list = []
    for i in range(0, len(session.index), shift):
        start = session.index[i]
        end = session.index[i] + (seq_length - 1) * pd.to_timedelta(freq)
        ex_sequence = session.loc[start:end, :]

        if len(ex_sequence.index) == seq_length:
            sequence_list.append(ex_sequence)
    return sequence_list


def infer_schema(schema):
    """
    Function to infer a schema from input data.
    Args:
        schema: Input dict mapping features to tf.Tensors.

    Returns:
        schema_dict: Dict mapping features to their respective data types.
    """
    schema_dict = dict()
    for feature, descriptions in schema.items():
        dtype = descriptions.dtype
        # These are ugly hacks for hard checking whether its broadly int, float
        #  bool or string.
        if dtype == float or 'float' in str(dtype):
            schema_dict[feature] = DefaultKeys.FLOAT
        elif dtype == int or 'int' in str(dtype):
            schema_dict[feature] = DefaultKeys.INTEGER
        elif dtype == bool or 'bool' in str(dtype):
            schema_dict[feature] = DefaultKeys.BOOLEAN
        elif dtype == str or 'string' in str(dtype):
            schema_dict[feature] = DefaultKeys.STRING
        else:
            raise AssertionError('Cannot infer schema dtype!')
    return schema_dict


def fill_defaults(config, default_config, schema):
    for feature in schema:
        if feature not in config:
            config[feature] = default_config[schema[feature]]
    return config


def get_function_dict(config, methods):
    result = dict()
    for feature, descriptions in config.items():
        result[feature] = []
        for d in descriptions:
            method = d[MethodKeys.METHOD]
            params = d[MethodKeys.PARAMETERS]

            result[feature].append(methods.get_method(method)(**params))

    return result



from typing import Text, Dict, Any, Optional

from tfx.dsl.components.base.base_component import BaseComponent
from tfx.dsl.components.base.executor_spec import ExecutorClassSpec
from tfx.types import standard_artifacts, Channel
from tfx.types.component_spec import ComponentSpec, ExecutionParameter, \
    ChannelParameter

from zenml.components.data_gen.constants import DATA_SPLIT_NAME
from zenml.components.data_gen.executor import DataExecutor
from zenml.standards.standard_keys import StepKeys


class DataGenSpec(ComponentSpec):
    PARAMETERS = {
        StepKeys.NAME: ExecutionParameter(type=Text),
        StepKeys.SOURCE: ExecutionParameter(type=Text),
        StepKeys.ARGS: ExecutionParameter(type=Dict[Text, Any]),
    }
    INPUTS = {}
    OUTPUTS = {
        DATA_SPLIT_NAME: ChannelParameter(type=standard_artifacts.Examples)
    }


class DataGen(BaseComponent):
    SPEC_CLASS = DataGenSpec
    EXECUTOR_SPEC = ExecutorClassSpec(DataExecutor)

    def __init__(self,
                 name: Text,
                 source: Text,
                 source_args: Dict[Text, Any],
                 instance_name: Optional[Text] = None,
                 examples: Optional[ChannelParameter] = None):
        """
        Interface for all DataGen components, the main component responsible
        for reading data and converting to TFRecords. This is how we handle
        versioning data for now.

        Args:
            name: name of datasource.
            source:
            source_args:
            instance_name:
            examples:
        """
        examples = examples or Channel(type=standard_artifacts.Examples)

        # Initiate the spec and create instance
        spec = self.SPEC_CLASS(name=name,
                               source=source,
                               args=source_args,
                               examples=examples)

        super(DataGen, self).__init__(spec=spec,
                                      instance_name=instance_name)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base Step Interface definition"""
import inspect
from typing import Dict

from zenml.backends import BaseBackend
from zenml.enums import StepTypes
from zenml.standards.standard_keys import StepKeys
from zenml.utils.print_utils import to_pretty_string, PrintStyles
from zenml.utils.source_utils import resolve_class, \
    load_source_path_class, is_valid_source


class BaseStep:
    """
    Base class for all ZenML steps.

    These are 'windows' into the base components for simpler overrides.
    """
    STEP_TYPE = StepTypes.base.name

    def __init__(self, backend: BaseBackend = None, **kwargs):
        """
        Base data step constructor.

        Args:
            **kwargs: Keyword arguments used in the construction of a step
            from a configuration file.
        """
        self._kwargs = kwargs
        self._immutable = False
        self.backend = backend
        self._source = resolve_class(self.__class__)

    def __str__(self):
        return to_pretty_string(self.to_config())

    def __repr__(self):
        return to_pretty_string(self.to_config(), style=PrintStyles.PPRINT)

    @staticmethod
    def from_config(config_block: Dict):
        """
        Takes config block that represents a Step and converts it back into
        its Python equivalent. This functionality is similar for most steps,
        and expected config_block may look like

        {
            'source': this.module.StepClass@sha  # where sha is optional
            'args': {}  # to be passed to the constructor
        }

        Args:
            config_block: config block representing source and args of step.
        """
        # resolve source path
        if StepKeys.SOURCE in config_block:
            source = config_block[StepKeys.SOURCE]
            class_ = load_source_path_class(source)
            args = config_block[StepKeys.ARGS]
            resolved_args = {}

            # resolve backend
            backend = None
            if StepKeys.BACKEND in config_block:
                backend_config = config_block[StepKeys.BACKEND]
                backend = BaseBackend.from_config(backend_config)

            # resolve args for special cases
            for k, v in args.items():
                if isinstance(v, str) and is_valid_source(v):
                    resolved_args[k] = load_source_path_class(v)
                else:
                    resolved_args[k] = v

            obj = class_(**resolved_args)

            # If we load from config, its immutable
            obj._immutable = True
            obj.backend = backend
            return obj
        else:
            raise AssertionError("Cannot create config_block without source "
                                 "key.")

    def to_config(self) -> Dict:
        """
        Converts from step back to config. This functionality for most steps
        yields the following config block.

        {
            'source': this.file.path.BaseStep@sha  # where sha is optional
            'args': {}  # whatever is used in the constructor for BaseStep
        }
        """
        kwargs = {}
        for key, kwarg in self._kwargs.items():
            if inspect.isclass(kwarg):
                kwargs[key] = resolve_class(kwarg)
            else:
                kwargs[key] = kwarg

        config = {
            StepKeys.SOURCE: self._source,
            StepKeys.ARGS: kwargs,  # everything to be recorded
        }

        # only add backend if its set
        if self.backend is not None:
            config.update({StepKeys.BACKEND: self.backend.to_config()})

        return config

    def with_backend(self, backend: BaseBackend):
        """Builder for step backends."""
        self.backend = backend
        return self



from typing import Any, Dict, List, Text, Callable, Optional

import apache_beam as beam
import tensorflow_model_analysis as tfma
from absl import logging
from tensorflow_model_analysis import constants as tfma_constants
from tfx import types
from tfx.components.util import tfxio_utils
from tfx.dsl.components.base import base_executor
from tfx.types import artifact_utils
from tfx.utils import import_utils
from tfx.utils import io_utils
from tfx.utils import path_utils
from tfx_bsl.tfxio import tensor_adapter

from zenml.components.evaluator import constants
from zenml.logger import get_logger
from zenml.standards.standard_keys import StepKeys
from zenml.steps.evaluator.base_evaluator import BaseEvaluatorStep
from zenml.steps.trainer.utils import TEST_SPLITS
from zenml.utils import path_utils as zenml_path_utils
from zenml.utils import source_utils

logger = get_logger(__name__)


def try_get_fn(module_path: Text,
               fn_name: Text) -> Optional[Callable[..., Any]]:
    try:
        return import_utils.import_func_from_module(module_path, fn_name)
    except (ValueError, AttributeError):
        return None


_TELEMETRY_DESCRIPTORS = ['Evaluator']


class Executor(base_executor.BaseExecutor):

    def Do(self,
           input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:

        # Check the inputs
        if constants.EXAMPLES not in input_dict:
            raise ValueError(f'{constants.EXAMPLES} is missing from inputs')
        examples_artifact = input_dict[constants.EXAMPLES]

        # Create the step with the schema attached if provided
        source = exec_properties[StepKeys.SOURCE]
        args = exec_properties[StepKeys.ARGS]
        c = source_utils.load_source_path_class(source)
        evaluator_step: BaseEvaluatorStep = c(**args)

        splits = evaluator_step.split_mapping[TEST_SPLITS]

        if len(splits) == 0:
            raise AssertionError(
                'In the current configuration of the evaluator step or the '
                'split mapping, there are no splits dedicated for the '
                'evaluation. Either use a "test" split or define a dict '
                'for the split mapping')
        else:
            # Check the outputs
            if constants.EVALUATION not in output_dict:
                raise ValueError(
                    f'{constants.EVALUATION} is missing from outputs')
            evaluation_artifact = output_dict[constants.EVALUATION]
            output_uri = artifact_utils.get_single_uri(evaluation_artifact)

            # Resolve the schema
            schema = None
            if constants.SCHEMA in input_dict:
                schema_artifact = input_dict[constants.SCHEMA]
                schema_uri = artifact_utils.get_single_uri(schema_artifact)
                reader = io_utils.SchemaReader()
                schema = reader.read(io_utils.get_only_uri_in_dir(schema_uri))

            # Check the execution parameters
            eval_config = evaluator_step.build_config()
            eval_config = tfma.update_eval_config_with_defaults(eval_config)
            tfma.verify_eval_config(eval_config)

            # Resolve the model
            if constants.MODEL in input_dict:
                model_artifact = input_dict[constants.MODEL]
                model_uri = artifact_utils.get_single_uri(model_artifact)
                model_path = path_utils.serving_model_path(model_uri)

                model_fn = try_get_fn(evaluator_step.CUSTOM_MODULE,
                                      'custom_eval_shared_model') or tfma.default_eval_shared_model

                eval_shared_model = model_fn(
                    model_name='',  # TODO: Fix with model names
                    eval_saved_model_path=model_path,
                    eval_config=eval_config)
            else:
                eval_shared_model = None

            self._log_startup(input_dict, output_dict, exec_properties)

            # Main pipeline
            logging.info('Evaluating model.')

            with self._make_beam_pipeline() as pipeline:
                examples_list = []
                tensor_adapter_config = None
                if tfma.is_batched_input(eval_shared_model, eval_config):
                    tfxio_factory = tfxio_utils.get_tfxio_factory_from_artifact(
                        examples=[
                            artifact_utils.get_single_instance(
                                examples_artifact)],
                        telemetry_descriptors=_TELEMETRY_DESCRIPTORS,
                        schema=schema,
                        raw_record_column_name=tfma_constants.ARROW_INPUT_COLUMN)

                    for split in splits:
                        split_uri = artifact_utils.get_split_uri(
                            examples_artifact,
                            split)

                        try:
                            if len(zenml_path_utils.list_dir(split_uri)) == 0:
                                raise AssertionError(
                                    f'The saved results found for the split: '
                                    f'{split}. Please make sure that you '
                                    f'run the test_fn in your trainer and '
                                    f'you are working with the right '
                                    f'split_mapping: '
                                    f'{evaluator_step.split_mapping}!')
                        except:
                            # TODO[LOW]: Merge into a single check
                            raise AssertionError(
                                f'The saved results found for the split: '
                                f'{split}. Please make sure that you '
                                f'run the test_fn in your trainer and '
                                f'you are working with the right '
                                f'split_mapping: '
                                f'{evaluator_step.split_mapping}!')

                        file_pattern = io_utils.all_files_pattern(split_uri)
                        tfxio = tfxio_factory(file_pattern)
                        data = (pipeline
                                | 'ReadFromTFRecordToArrow[%s]' % split >> tfxio.BeamSource())
                        examples_list.append(data)
                    if schema is not None:
                        tensor_adapter_config = tensor_adapter.TensorAdapterConfig(
                            arrow_schema=tfxio.ArrowSchema(),
                            tensor_representations=tfxio.TensorRepresentations())
                else:
                    for split in splits:
                        file_pattern = io_utils.all_files_pattern(
                            artifact_utils.get_split_uri(examples_artifact,
                                                         split))
                        data = (pipeline
                                | 'ReadFromTFRecord[%s]' % split >>
                                beam.io.ReadFromTFRecord(
                                    file_pattern=file_pattern)
                                )
                        examples_list.append(data)
                # Resolve custom extractors
                custom_extractors = try_get_fn(evaluator_step.CUSTOM_MODULE,
                                               'custom_extractors')
                extractors = None
                if custom_extractors:
                    extractors = custom_extractors(
                        eval_shared_model=eval_shared_model,
                        eval_config=eval_config,
                        tensor_adapter_config=tensor_adapter_config)

                # Resolve custom evaluators
                custom_evaluators = try_get_fn(evaluator_step.CUSTOM_MODULE,
                                               'custom_evaluators')
                evaluators = None
                if custom_evaluators:
                    evaluators = custom_evaluators(
                        eval_shared_model=eval_shared_model,
                        eval_config=eval_config,
                        tensor_adapter_config=tensor_adapter_config)

                # Extract, evaluate and write
                (examples_list | 'FlattenExamples' >> beam.Flatten()
                 | 'ExtractEvaluateAndWriteResults' >> tfma.ExtractEvaluateAndWriteResults(
                            eval_config=eval_config,
                            eval_shared_model=eval_shared_model,
                            output_path=output_uri,
                            extractors=extractors,
                            evaluators=evaluators,
                            tensor_adapter_config=tensor_adapter_config))
            logging.info('Evaluation complete. Results written to %s.',
                         output_uri)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import click

from zenml.cli.cli import cli
from zenml.version import __version__


@cli.command()
def version():
    """Version of ZenML"""
    click.echo(click.style(r"""      
           .-') _   ('-.       .-') _  _   .-')              
          (  OO) )_(  OO)     ( OO ) )( '.( OO )_            
        ,(_)----.(,------.,--./ ,--,'  ,--.   ,--.),--.      
        |       | |  .---'|   \ |  |\  |   `.'   | |  |.-')  
        '--.   /  |  |    |    \|  | ) |         | |  | OO ) 
        (_/   /  (|  '--. |  .     |/  |  |'.'|  | |  |`-' | 
         /   /___ |  .--' |  |\    |   |  |   |  |(|  '---.' 
        |        ||  `---.|  | \   |   |  |   |  | |      |  
        `--------'`------'`--'  `--'   `--'   `--' `------' 
         """, fg='green'))
    click.echo(click.style(f"version: {__version__}", bold=True))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict, Text, Any, List
from typing import Optional

from tfx.components.common_nodes.importer_node import ImporterNode
from tfx.components.pusher.component import Pusher
from tfx.types import standard_artifacts

from zenml.backends.orchestrator import OrchestratorBaseBackend
from zenml.datasources import BaseDatasource
from zenml.metadata import ZenMLMetadataStore
from zenml.pipelines import BasePipeline
from zenml.repo import ArtifactStore
from zenml.standards import standard_keys as keys
from zenml.steps import BaseStep
from zenml.steps.deployer import BaseDeployerStep
from zenml.enums import GDPComponent


class DeploymentPipeline(BasePipeline):
    """BatchInferencePipeline definition to run batch inference pipelines.

    A BatchInferencePipeline is used to run an inference based on a
    TrainingPipeline.
    """
    PIPELINE_TYPE = 'deploy'

    def __init__(self,
                 model_uri: Text,
                 name: Text = None,
                 enable_cache: Optional[bool] = True,
                 steps_dict: Dict[Text, BaseStep] = None,
                 backend: OrchestratorBaseBackend = None,
                 metadata_store: Optional[ZenMLMetadataStore] = None,
                 artifact_store: Optional[ArtifactStore] = None,
                 datasource: Optional[BaseDatasource] = None,
                 pipeline_name: Optional[Text] = None):
        """
        Construct a deployment pipeline. This is a pipeline that deploys a
        model to a target. Can be controlled via a DeployerStep.

        Args:
            name: Outward-facing name of the pipeline.
            model_uri: URI for a model, usually generated by
            TrainingPipeline and retrieved by
            `training_pipeline.get_model_uri()`.
            pipeline_name: A unique name that identifies the pipeline after
             it is run.
            enable_cache: Boolean, indicates whether or not caching
             should be used.
            steps_dict: Optional dict of steps.
            backend: Orchestrator backend.
            metadata_store: Configured metadata store. If None,
             the default metadata store is used.
            artifact_store: Configured artifact store. If None,
             the default artifact store is used.
        """
        if model_uri is None:
            raise AssertionError('model_uri cannot be None.')
        self.model_uri = model_uri
        super(DeploymentPipeline, self).__init__(
            name=name,
            enable_cache=enable_cache,
            steps_dict=steps_dict,
            backend=backend,
            metadata_store=metadata_store,
            artifact_store=artifact_store,
            datasource=datasource,
            pipeline_name=pipeline_name,
            model_uri=model_uri,
        )

    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Creates an inference pipeline out of TFX components.

        A inference pipeline is used to run a batch of data through a
        ML model via the BulkInferrer TFX component.

        Args:
            config: Dict. Contains a ZenML configuration used to build the
             data pipeline.

        Returns:
            A list of TFX components making up the data pipeline.
        """
        component_list = []

        # Load from model_uri
        model = ImporterNode(
            instance_name=GDPComponent.Trainer.name,
            source_uri=self.model_uri,
            artifact_type=standard_artifacts.Model)
        model_result = model.outputs.result

        deployer: BaseDeployerStep = \
            self.steps_dict[keys.TrainingSteps.DEPLOYER]
        pusher_config = deployer._build_pusher_args()
        pusher_executor_spec = deployer._get_executor_spec()
        pusher = Pusher(model_export=model_result,
                        custom_executor_spec=pusher_executor_spec,
                        **pusher_config).with_id(
            GDPComponent.Deployer.name)

        component_list.extend([model, pusher])
        return component_list

    def add_deployment(self, deployment_step: BaseDeployerStep):
        self.steps_dict[keys.TrainingSteps.DEPLOYER] = deployment_step

    def steps_completed(self) -> bool:
        mandatory_steps = [keys.TrainingSteps.DEPLOYER]
        for step_name in mandatory_steps:
            if step_name not in self.steps_dict.keys():
                raise AssertionError(
                    f'Mandatory step {step_name} not added.')
        return True



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Training pipeline step to create a pipeline that trains on data."""

import json
import os
from pathlib import Path
from typing import Dict, Text, Any, List

import tensorflow as tf
from tfx.components.common_nodes.importer_node import ImporterNode
from tfx.components.pusher.component import Pusher
from tfx.components.schema_gen.component import SchemaGen
from tfx.components.statistics_gen.component import StatisticsGen
from tfx.components.transform.component import Transform
from tfx.types import standard_artifacts

from zenml import constants
from zenml.backends.training import TrainingBaseBackend
from zenml.components import Sequencer, SplitGen, Trainer, Evaluator
from zenml.enums import GDPComponent
from zenml.exceptions import DoesNotExistException, \
    PipelineNotSucceededException
from zenml.logger import get_logger
from zenml.pipelines import BasePipeline
from zenml.standards import standard_keys as keys
from zenml.steps.deployer import BaseDeployerStep
from zenml.steps.evaluator import BaseEvaluatorStep
from zenml.steps.preprocesser import BasePreprocesserStep
from zenml.steps.preprocesser.base_preprocesser import build_split_mapping
from zenml.steps.sequencer import BaseSequencerStep
from zenml.steps.split import BaseSplitStep
from zenml.steps.trainer import BaseTrainerStep
from zenml.utils import path_utils
from zenml.utils.post_training.post_training_utils import \
    evaluate_single_pipeline, view_statistics, view_schema, detect_anomalies
from zenml.utils.post_training.post_training_utils import \
    get_feature_spec_from_schema, convert_raw_dataset_to_pandas

logger = get_logger(__name__)


class TrainingPipeline(BasePipeline):
    """
    Definition of the Training Pipeline class.

    TrainingPipeline is a general-purpose training pipeline for most ML
    training runs. One pipeline runs one experiment on a single datasource.
    """

    PIPELINE_TYPE = 'training'

    def get_tfx_component_list(self, config: Dict[Text, Any]) -> List:
        """
        Builds the training pipeline as a series of TFX components.

        Args:
            config: A ZenML configuration in dictionary format.

        Returns:
            A chronological list of TFX components making up the training
             pipeline.

        """
        steps = config[keys.GlobalKeys.PIPELINE][keys.PipelineKeys.STEPS]

        component_list = []

        ############
        # RAW DATA #
        ############
        # if there are no commits, then lets make one
        if self.datasource.is_empty:
            logger.info(
                f'Datasource {self.datasource.name} has no commits. Creating '
                f'the first one..')
            self.datasource_commit_id = self.datasource.commit()

        data_pipeline = self.datasource.get_data_pipeline_from_commit(
            self.datasource_commit_id)

        data = ImporterNode(
            instance_name=GDPComponent.DataGen.name,
            source_uri=data_pipeline.get_artifacts_uri_by_component(
                GDPComponent.DataGen.name)[0],
            artifact_type=standard_artifacts.Examples)

        schema_data = ImporterNode(
            instance_name=GDPComponent.DataSchema.name,
            source_uri=data_pipeline.get_artifacts_uri_by_component(
                GDPComponent.DataSchema.name)[0],
            artifact_type=standard_artifacts.Schema)

        statistics_data = ImporterNode(
            instance_name=GDPComponent.DataStatistics.name,
            source_uri=data_pipeline.get_artifacts_uri_by_component(
                GDPComponent.DataStatistics.name)[0],
            artifact_type=standard_artifacts.ExampleStatistics)

        component_list.extend([data, schema_data, statistics_data])

        datapoints = data.outputs.result

        #############
        # SPLITTING #
        #############
        # Block to read the data from the corresponding BQ table
        split_config = steps[keys.TrainingSteps.SPLIT]
        splits = SplitGen(
            input_examples=datapoints,
            source=split_config[keys.StepKeys.SOURCE],
            source_args=split_config[keys.StepKeys.ARGS],
            schema=schema_data.outputs.result,
            statistics=statistics_data.outputs.result,
        ).with_id(GDPComponent.SplitGen.name)

        datapoints = splits.outputs.examples

        statistics_split = StatisticsGen(
            examples=datapoints
        ).with_id(GDPComponent.SplitStatistics.name)

        schema_split = SchemaGen(
            statistics=statistics_split.outputs.output,
        ).with_id(GDPComponent.SplitSchema.name)

        schema = schema_split.outputs.schema

        component_list.extend([splits,
                               statistics_split,
                               schema_split])

        ##############
        # SEQUENCING #
        ##############
        if keys.TrainingSteps.SEQUENCER in steps:
            sequencer_config = steps[keys.TrainingSteps.SEQUENCER]
            sequencer = Sequencer(
                input_examples=datapoints,
                schema=schema,
                statistics=statistics_split.outputs.statistics,
                source=sequencer_config[keys.StepKeys.SOURCE],
                source_args=sequencer_config[keys.StepKeys.ARGS]
            ).with_id(GDPComponent.Sequencer.name)

            sequencer_statistics = StatisticsGen(
                examples=sequencer.outputs.output_examples
            ).with_id(GDPComponent.SequencerStatistics.name)

            sequencer_schema = SchemaGen(
                statistics=sequencer_statistics.outputs.output,
                infer_feature_shape=True,
            ).with_id(GDPComponent.SequencerSchema.name)

            datapoints = sequencer.outputs.output_examples
            schema = sequencer_schema.outputs.schema

            component_list.extend([sequencer,
                                   sequencer_statistics,
                                   sequencer_schema])

        #################
        # PREPROCESSING #
        #################
        split_mapping = build_split_mapping(
            steps[keys.TrainingSteps.PREPROCESSER][keys.StepKeys.ARGS])

        transform = Transform(
            preprocessing_fn=constants.PREPROCESSING_FN,
            examples=datapoints,
            schema=schema,
            splits_config=split_mapping,
            custom_config=steps[keys.TrainingSteps.PREPROCESSER]
        ).with_id(GDPComponent.Transform.name)

        component_list.extend([transform])

        ############
        # TRAINING #
        ############
        training_backend: TrainingBaseBackend = \
            self.steps_dict[keys.TrainingSteps.TRAINER].backend

        # default to local
        if training_backend is None:
            training_backend = TrainingBaseBackend()

        training_kwargs = {
            'custom_executor_spec': training_backend.get_executor_spec(),
            'custom_config': steps[keys.TrainingSteps.TRAINER]
        }
        training_kwargs['custom_config'].update(
            training_backend.get_custom_config())

        trainer = Trainer(
            transformed_examples=transform.outputs.transformed_examples,
            transform_graph=transform.outputs.transform_graph,
            run_fn=constants.TRAINER_FN,
            schema=schema,
            **training_kwargs
        ).with_id(GDPComponent.Trainer.name)

        component_list.extend([trainer])

        #############
        # EVALUATOR #
        #############
        if keys.TrainingSteps.EVALUATOR in steps:
            evaluator_config = steps[keys.TrainingSteps.EVALUATOR]
            # TODO [MEDIUM]: This check should ideally not happen here,
            #  integrate it into the component
            eval_source = evaluator_config['source'].split('@')[0]
            if eval_source == \
                    'zenml.steps.evaluator.agnostic_evaluator' \
                    '.AgnosticEvaluator':
                evaluator = Evaluator(
                    source=evaluator_config[keys.StepKeys.SOURCE],
                    source_args=evaluator_config[keys.StepKeys.ARGS],
                    examples=trainer.outputs.test_results,
                ).with_id(GDPComponent.Evaluator.name)
            elif eval_source == \
                    'zenml.steps.evaluator.tfma_evaluator.TFMAEvaluator':
                evaluator = Evaluator(
                    source=evaluator_config[keys.StepKeys.SOURCE],
                    source_args=evaluator_config[keys.StepKeys.ARGS],
                    examples=transform.outputs.transformed_examples,
                    model=trainer.outputs.model,
                ).with_id(GDPComponent.Evaluator.name)
            else:
                raise ValueError(
                    'Please use either the built-in TFMAEvaluator '
                    'or the AgnosticEvaluator')
            component_list.append(evaluator)

        ###########
        # SERVING #
        ###########
        if keys.TrainingSteps.DEPLOYER in steps:
            deployer: BaseDeployerStep = \
                self.steps_dict[keys.TrainingSteps.DEPLOYER]
            pusher_config = deployer._build_pusher_args()
            pusher_executor_spec = deployer._get_executor_spec()
            pusher = Pusher(model_export=trainer.outputs.model,
                            custom_executor_spec=pusher_executor_spec,
                            **pusher_config).with_id(
                GDPComponent.Deployer.name)

            component_list.append(pusher)

        return component_list

    def add_split(self, split_step: BaseSplitStep):
        self.steps_dict[keys.TrainingSteps.SPLIT] = split_step

    def add_sequencer(self, sequencer_step: BaseSequencerStep):
        self.steps_dict[keys.TrainingSteps.SEQUENCER] = sequencer_step

    def add_preprocesser(self, preprocessor_step: BasePreprocesserStep):
        self.steps_dict[keys.TrainingSteps.PREPROCESSER] = preprocessor_step

    def add_trainer(self, trainer_step: BaseTrainerStep):
        self.steps_dict[keys.TrainingSteps.TRAINER] = trainer_step

    def add_evaluator(self, evaluator_step: BaseEvaluatorStep):
        self.steps_dict[keys.TrainingSteps.EVALUATOR] = evaluator_step

    def add_deployment(self, deployment_step: BaseDeployerStep):
        self.steps_dict[keys.TrainingSteps.DEPLOYER] = deployment_step

    def view_statistics(self, magic: bool = False, port: int = 0):
        """
        View statistics for training pipeline in HTML.

        Args:
            magic (bool): Creates HTML page if False, else
            creates a notebook cell.
            port (int): Port at which to launch the statistics facet.
        """
        logger.info(
            'Viewing statistics. If magic=False then a new window will open '
            'up with a notebook for evaluation. If magic=True, then an '
            'attempt will be made to append to the current notebook.')
        uri = self.get_artifacts_uri_by_component(
            GDPComponent.SplitStatistics.name)[0]
        view_statistics(uri, magic, port)

    def view_schema(self):
        """View schema of data flowing in pipeline."""
        uri = self.get_artifacts_uri_by_component(
            GDPComponent.SplitSchema.name)[0]
        view_schema(uri)

    def evaluate(self, magic: bool = False, port: int = 0):
        """
        Evaluate pipeline from the evaluator and trainer steps artifact.

        Args:
            magic: Creates new window if False, else creates notebook cells.
            port: At which port to deploy jupyter notebook.
        """
        from zenml.enums import PipelineStatusTypes
        if self.get_status() != PipelineStatusTypes.Succeeded.name:
            logger.info(
                "Cannot evaluate a pipeline that has not executed "
                "successfully. Please run the pipeline and ensure it "
                "completes successfully to evaluate it.")
            raise PipelineNotSucceededException(name=self.name)
        if keys.TrainingSteps.EVALUATOR not in self.steps_dict:
            logger.info("This pipeline does not contain an Evaluator step.")
            raise DoesNotExistException(
                name=f'{keys.TrainingSteps.EVALUATOR}',
                reason='This pipeline does not contain an Evaluator step!')
        if keys.TrainingSteps.TRAINER not in self.steps_dict:
            logger.info("This pipeline does not contain a TRAINER step.")
            raise DoesNotExistException(
                name=f'{keys.TrainingSteps.TRAINER}',
                reason='This pipeline does not contain a Trainer step!')

        logger.info(
            'Evaluating pipeline. If magic=False then a new window will open '
            'up with a notebook for evaluation. If magic=True, then an '
            'attempt will be made to append to the current notebook.')
        return evaluate_single_pipeline(self, magic=magic, port=port)

    def download_model(self, out_path: Text = None, overwrite: bool = False):
        """Download model to out_path"""
        # TODO: [LOW] Implement when Deployer logic figured out
        raise NotImplementedError('Not implemented!')
        # model_path = self.get_artifacts_uri_by_component(
        #     GDPComponent.Deployer.name)
        # if out_path:
        #     from zenml.utils.path_utils import move
        #     move(model_path, out_path, overwrite)
        # else:
        #     out_path = model_path
        # logger.debug(f'Model: {out_path}')

    def view_anomalies(self, split_name='eval'):
        """
        View schema of data flowing in pipeline.

        Args:
            split_name: name of split to detect anomalies on
        """
        stats_uri = self.get_artifacts_uri_by_component(
            GDPComponent.SplitStatistics.name)[0]
        schema_uri = self.get_artifacts_uri_by_component(
            GDPComponent.SplitSchema.name)[0]
        detect_anomalies(stats_uri, schema_uri, split_name)

    def steps_completed(self) -> bool:
        mandatory_steps = [keys.TrainingSteps.SPLIT,
                           keys.TrainingSteps.PREPROCESSER,
                           keys.TrainingSteps.TRAINER]
        for step_name in mandatory_steps:
            if step_name not in self.steps_dict.keys():
                raise AssertionError(f'Mandatory step {step_name} not added.')
        return True

    def get_model_uri(self):
        """Gets model artifact."""
        uris = self.get_artifacts_uri_by_component(
            GDPComponent.Trainer.name, False)
        return uris[0]

    def get_hyperparameters(self) -> Dict:
        """
        Gets all hyper-parameters of pipeline.
        """
        executions = self.metadata_store.get_pipeline_executions(self)

        hparams = {}
        for e in executions:
            component_id = e.properties['component_id'].string_value
            if component_id == GDPComponent.Trainer.name:
                custom_config = json.loads(
                    e.properties['custom_config'].string_value)
                fn = custom_config[keys.StepKeys.SOURCE]
                params = custom_config[keys.StepKeys.ARGS]
                # filter out None values to not break compare tool
                params = {k: v for k, v in params.items() if v is not None}
                hparams[f'{component_id}_fn'] = fn
                hparams.update(params)
        return hparams

    def sample_transformed_data(self,
                                split_name: Text = 'eval',
                                sample_size: int = 100000):
        """
        Samples transformed data as a pandas DataFrame.

        Args:
            split_name: name of split to see
            sample_size: # of rows to sample.
        """
        base_uri = self.get_artifacts_uri_by_component(
            GDPComponent.Transform.name)[0]
        transform_schema = os.path.join(base_uri, 'transformed_metadata')
        spec = get_feature_spec_from_schema(transform_schema)

        base_uri = Path(base_uri)
        id_ = base_uri.name
        transform_data_path = os.path.join(
            str(base_uri.parent.parent), 'transformed_examples', id_)

        split_data_path = os.path.join(transform_data_path, split_name)
        data_files = path_utils.list_dir(split_data_path)
        dataset = tf.data.TFRecordDataset(data_files, compression_type='GZIP')
        return convert_raw_dataset_to_pandas(dataset, spec, sample_size)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
import shutil
from pathlib import Path

import pytest

import zenml
from zenml.metadata import MockMetadataStore
from zenml.repo.constants import ARTIFACT_STORE_DEFAULT_DIR, \
    ZENML_DIR_NAME, ML_METADATA_SQLITE_DEFAULT_NAME
from zenml.repo.zenml_config import PIPELINES_DIR_KEY
from zenml.repo import ZenMLConfig
from zenml.standards import standard_keys as keys
from zenml.utils import yaml_utils
from zenml.exceptions import InitializationException

# Nicholas a way to get to the root
ZENML_ROOT = str(Path(zenml.__path__[0]).parent)
TEST_ROOT = os.path.join(ZENML_ROOT, "tests")

pipelines_dir = os.path.join(TEST_ROOT, "pipelines")

config_root = os.path.dirname(TEST_ROOT)
artifact_store_path = os.path.join(config_root, ZENML_DIR_NAME,
                                   ARTIFACT_STORE_DEFAULT_DIR)

sqlite_uri = os.path.join(artifact_store_path, ML_METADATA_SQLITE_DEFAULT_NAME)


def test_zenml_config_init():
    # in the root initialization should work
    _ = ZenMLConfig(TEST_ROOT)

    # outside of an initialized repo path
    with pytest.raises(InitializationException):
        _ = ZenMLConfig(pipelines_dir)


def test_is_zenml_dir():
    ok_path = TEST_ROOT

    not_ok_path = pipelines_dir

    assert ZenMLConfig.is_zenml_dir(ok_path)
    assert not ZenMLConfig.is_zenml_dir(not_ok_path)


# def test_to_from_config(equal_zenml_configs):
#     # TODO: This is messed up
#     cfg1 = ZenMLConfig(repo_path=TEST_ROOT)
#
#     new_zenml_dir = os.path.join(TEST_ROOT, ".zenml")
#
#     os.mkdir(new_zenml_dir)
#
#     cfg2 = cfg1.from_config(cfg1.to_config(path=TEST_ROOT))
#
#     assert equal_zenml_configs(cfg1, cfg2, loaded=True)
#
#     shutil.rmtree(new_zenml_dir, ignore_errors=False)


def test_zenml_config_getters():
    cfg1 = ZenMLConfig(repo_path=TEST_ROOT)

    assert cfg1.get_pipelines_dir()
    assert cfg1.get_artifact_store()
    assert cfg1.get_metadata_store()


def test_zenml_config_setters(equal_md_stores):
    cfg1 = ZenMLConfig(repo_path=TEST_ROOT)

    old_store_path = artifact_store_path
    old_pipelines_dir = pipelines_dir
    old_sqlite = cfg1.get_metadata_store()

    new_store_path = os.getcwd()
    new_pipelines_dir = "awfkoeghelk"
    new_mdstore = MockMetadataStore()

    cfg1.set_artifact_store(new_store_path)
    cfg1.set_pipelines_dir(new_pipelines_dir)
    cfg1.set_metadata_store(new_mdstore)

    updated_cfg = yaml_utils.read_yaml(cfg1.config_path)

    loaded_md_store = MockMetadataStore.from_config(
        updated_cfg[keys.GlobalKeys.METADATA_STORE])

    assert updated_cfg[keys.GlobalKeys.ARTIFACT_STORE] == new_store_path
    assert updated_cfg[PIPELINES_DIR_KEY] == new_pipelines_dir

    assert equal_md_stores(new_mdstore, loaded_md_store)

    # revert changes
    cfg1.set_artifact_store(old_store_path)
    cfg1.set_pipelines_dir(old_pipelines_dir)
    cfg1.set_metadata_store(old_sqlite)

    shutil.rmtree(new_pipelines_dir, ignore_errors=False)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.


from typing import Dict, List, Text

import tensorflow_model_analysis as tfma
from google.protobuf.wrappers_pb2 import BoolValue

from zenml.steps.evaluator.base_evaluator import BaseEvaluatorStep


def to_camel_case(s):
    """
    Converts a given name to camel case and removes any '_'
    """
    return ''.join(list(map(lambda x: x.capitalize(), s.split('_'))))


class TFMAEvaluator(BaseEvaluatorStep):
    """
    TFMA Evaluator step designed for the TF models in ZenML

    It features a specific build_config method which uses the zenml_eval
    signature of the model and produces a flexible tfma.EvalConfig, which can
    work with a wide range of Tensorflow models and TFMA
    """

    CUSTOM_MODULE = 'zenml.steps.evaluator.tfma_module'

    def __init__(self,
                 metrics: Dict[Text, List[Text]],
                 slices: List[List[Text]] = None,
                 output_mapping: Dict[Text, Text] = None,
                 **kwargs):
        """
        Init for the TFMA evaluator

        :param metrics: a dictionary, which specifies a list metrics for each
        label
        :param slices: a list of lists, each element in the inner list include
        a set of features which will be used for slicing on the results
        :param output_mapping: a mapping from label names to output names.
        This is especially useful, when it comes to multi-output models or
        models with unknown output names.
        :param splits: the list of splits to apply the evaluation on
        """

        super(TFMAEvaluator, self).__init__(slices=slices,
                                            metrics=metrics,
                                            output_mapping=output_mapping,
                                            **kwargs)

        self.metrics = metrics
        self.slices = slices or list()

        if output_mapping is None:
            if len(self.metrics.keys()) == 1:
                label = list(self.metrics.keys())[0]
                self.output_mapping = {label: label}
            else:
                raise ValueError('If you are using the evaluator to compute '
                                 'metrics on more than 1 output/label please'
                                 'provide a output_mapping!')
        else:
            self.output_mapping = output_mapping

    def build_config(self, use_defaults=False):
        # SLICING SPEC
        slicing_specs = [tfma.SlicingSpec()]
        if self.slices:
            slicing_specs.extend([tfma.SlicingSpec(feature_keys=e)
                                  for e in self.slices])

        # MODEL SPEC
        metric_labels = sorted(list(set(self.metrics.keys())))
        model_specs = [tfma.ModelSpec(signature_name='zen_eval',
                                      label_keys=self.output_mapping)]

        # METRIC SPEC
        baseline = [tfma.MetricConfig(class_name='ExampleCount')]
        metrics_specs = []
        for i, key in enumerate(metric_labels):
            metrics = baseline.copy()
            metrics.extend([tfma.MetricConfig(class_name=to_camel_case(m))
                            for m in self.metrics[key]])

            metrics_specs.append(tfma.MetricsSpec(
                output_names=[key],
                metrics=metrics))

        return tfma.EvalConfig(
            model_specs=model_specs,
            slicing_specs=slicing_specs,
            metrics_specs=metrics_specs,
            options=tfma.Options(
                include_default_metrics=BoolValue(value=use_defaults)))



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Definition of the GCAIP Training Backend"""

import time
from typing import Text

from tfx.dsl.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer.executor \
    import TRAINING_ARGS_KEY, JOB_ID_KEY

from zenml.backends.training import TrainingBaseBackend
from zenml.components.trainer.gcaip_executor import \
    ZenMLTrainerGCAIPExecutor
from zenml.constants import ZENML_TRAINER_IMAGE_NAME
from zenml.enums import GCPGPUTypes
from zenml.logger import get_logger
from zenml.pipelines.utils import sanitize_name_for_ai_platform
from zenml.utils import requirement_utils

requirement_utils.check_integration(requirement_utils.GCP_INTEGRATION)

logger = get_logger(__name__)


class SingleGPUTrainingGCAIPBackend(TrainingBaseBackend):
    """
    Runs a TrainerStep on Google Cloud AI Platform.

    A training backend can be used to efficiently train a machine learning
    model on large amounts of data. This triggers a Training job on the Google
    Cloud AI Platform service: https://cloud.google.com/ai-platform.

    This backend is meant for a training job with a single GPU only. The user
    has a choice of three GPUs, specified in the GCPGPUTypes Enum.
    """

    def __init__(
            self,
            project: Text,
            job_dir: Text,
            gpu_type: Text = GCPGPUTypes.K80.name,
            machine_type: Text = 'n1-standard-4',
            image: Text = ZENML_TRAINER_IMAGE_NAME,
            job_name: Text = f'train_{int(time.time())}',
            region: Text = 'europe-west1',
            python_version: Text = '3.7',
            max_running_time: int = 7200,
            **kwargs):
        """
        An opinionated wrapper around a GCAIP training job.

        Args:
            project: The GCP project in which to run the job.
            job_dir: A bucket where to store some metadata while training.
            gpu_type: The type of gpu.
            machine_type: The type of machine to use. This must conform to
            the GCP compatability matrix with the gpu_type. See details
            here: https://cloud.google.com/ai-platform/training/docs/using
            -gpus#compute-engine-machine-types-with-gpu
            image: The Docker image with which to run the job.
            job_name: The name of the job.
            region: The GCP region to run the job in.
            python_version: The Python version for the job.
            max_running_time: The maximum running time of the job in seconds.
            **kwargs:
        """
        if gpu_type not in GCPGPUTypes.__members__.keys():
            raise AssertionError(f'gpu_type must be one of: '
                                 f'{list(GCPGPUTypes.__members__.keys())}')
        self.project = project
        self.job_dir = job_dir
        self.gpu_type = gpu_type
        self.machine_type = machine_type
        self.image = image
        self.job_name = job_name
        self.region = region
        self.python_version = python_version
        self.max_running_time = max_running_time
        super().__init__(
            project=project,
            job_dir=job_dir,
            gpu_type=gpu_type,
            machine_type=machine_type,
            image=image,
            job_name=job_name,
            region=region,
            python_version=python_version,
            max_running_time=max_running_time,
            **kwargs
        )

    def get_executor_spec(self):
        return executor_spec.ExecutorClassSpec(ZenMLTrainerGCAIPExecutor)

    def get_custom_config(self):
        train_args = {
            'project': self.project,
            'region': self.region,
            'jobDir': self.job_dir,
            'masterConfig': {
                'imageUri': self.image,
                'acceleratorConfig': {
                    'count': 1,  # always 1 to avoid complications
                    'type': f'NVIDIA_TESLA_{self.gpu_type}'
                }
            },
            'masterType': self.machine_type,
            'scaleTier': 'CUSTOM',
            'pythonVersion': self.python_version,
            'scheduling': {'maxRunningTime': f'{self.max_running_time}s'},
        }
        logger.debug(f'Configuring GCAIP Trainer job {self.job_name} '
                     f'with args: {train_args}')
        logger.info(f'Launching GCAIP Trainer job {self.job_name}')
        return {
            TRAINING_ARGS_KEY: train_args,
            JOB_ID_KEY: sanitize_name_for_ai_platform(
                self.job_name)
        }



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text

from tfx.orchestration import metadata

from zenml.metadata import ZenMLMetadataStore
from zenml.utils import path_utils
from zenml.enums import MLMetadataTypes


class SQLiteMetadataStore(ZenMLMetadataStore):
    STORE_TYPE = MLMetadataTypes.sqlite.name

    def __init__(self, uri: Text):
        """Constructor for MySQL MetadataStore for ZenML"""
        if path_utils.is_remote(uri):
            raise Exception(f'URI {uri} is a non-local path. A sqlite store '
                            f'can only be local paths')

        # Resolve URI if relative URI provided
        self.uri = path_utils.resolve_relative_path(uri)
        super().__init__()

    def get_tfx_metadata_config(self):
        return metadata.sqlite_metadata_connection_config(self.uri)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import os
from typing import Any, Dict, List, Text

import apache_beam as beam
from apache_beam.transforms.window import GlobalWindows
from tensorflow_data_validation.coders import tf_example_decoder
from tfx import types
from tfx.dsl.components.base.base_executor import BaseExecutor
from tfx.types import artifact_utils
from tfx.utils import io_utils

from zenml.components.sequencer import constants, utils
from zenml.standards.standard_keys import StepKeys
from zenml.steps.sequencer import BaseSequencerStep
from zenml.utils import source_utils


class RemoveKey(beam.DoFn):
    def process(self, element):
        """
        Remove the key from the output of the CombineFn and return the
        resulting datapoints
        """
        return element[1]


class Executor(BaseExecutor):
    _DEFAULT_FILENAME = 'sequencer_output'

    def Do(self,
           input_dict: Dict[Text, List[types.Artifact]],
           output_dict: Dict[Text, List[types.Artifact]],
           exec_properties: Dict[Text, Any]) -> None:
        """
        Main execution logic for the Sequencer component

        :param input_dict: input channels
        :param output_dict: output channels
        :param exec_properties: the execution properties defined in the spec
        """

        source = exec_properties[StepKeys.SOURCE]
        args = exec_properties[StepKeys.ARGS]

        c = source_utils.load_source_path_class(source)

        # Get the schema
        schema_path = io_utils.get_only_uri_in_dir(
            artifact_utils.get_single_uri(input_dict[constants.SCHEMA]))
        schema = io_utils.SchemaReader().read(schema_path)

        # TODO: Getting the statistics might help the future implementations

        sequence_step: BaseSequencerStep = c(schema=schema,
                                             statistics=None,
                                             **args)

        # Get split names
        input_artifact = artifact_utils.get_single_instance(
            input_dict[constants.INPUT_EXAMPLES])
        split_names = artifact_utils.decode_split_names(
            input_artifact.split_names)

        # Create output artifact
        output_artifact = artifact_utils.get_single_instance(
            output_dict[constants.OUTPUT_EXAMPLES])
        output_artifact.split_names = artifact_utils.encode_split_names(
            split_names)

        with self._make_beam_pipeline() as p:
            for s in split_names:
                input_uri = io_utils.all_files_pattern(
                    artifact_utils.get_split_uri(
                        input_dict[constants.INPUT_EXAMPLES], s))

                output_uri = artifact_utils.get_split_uri(
                    output_dict[constants.OUTPUT_EXAMPLES],
                    s)
                output_path = os.path.join(output_uri, self._DEFAULT_FILENAME)

                # Read and decode the data
                data = \
                    (p
                     | 'Read_' + s >> beam.io.ReadFromTFRecord(
                                file_pattern=input_uri)
                     | 'Decode_' + s >> tf_example_decoder.DecodeTFExample()
                     | 'ToDataFrame_' + s >> beam.ParDo(
                                utils.ConvertToDataframe()))

                # Window into sessions
                s_data = \
                    (data
                     | 'AddCategory_' + s >> beam.ParDo(
                                sequence_step.get_category_do_fn())
                     | 'AddTimestamp_' + s >> beam.ParDo(
                                sequence_step.get_timestamp_do_fn())
                     | 'Sessions_' + s >> beam.WindowInto(
                                sequence_step.get_window()))

                # Combine and transform
                p_data = \
                    (s_data
                     | 'Combine_' + s >> beam.CombinePerKey(
                                sequence_step.get_combine_fn()))

                # Write the results
                _ = \
                    (p_data
                     | 'Global_' + s >> beam.WindowInto(GlobalWindows())
                     | 'RemoveKey_' + s >> beam.ParDo(RemoveKey())
                     | 'ToExample_' + s >> beam.Map(utils.df_to_example)
                     | 'Serialize_' + s >> beam.Map(utils.serialize)
                     | 'Write_' + s >> beam.io.WriteToTFRecord(
                                output_path,
                                file_name_suffix='.gz'))



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

import tensorflow as tf
import tensorflow_addons as tfa


class CycleGan(tf.keras.Model):
    def __init__(
            self,
            monet_generator: tf.keras.Model,
            photo_generator: tf.keras.Model,
            monet_discriminator: tf.keras.Model,
            photo_discriminator: tf.keras.Model,
            lambda_cycle: int = 10,
    ):
        super(CycleGan, self).__init__()
        self.m_gen = monet_generator
        self.p_gen = photo_generator
        self.m_disc = monet_discriminator
        self.p_disc = photo_discriminator
        self.lambda_cycle = lambda_cycle

        self.m_gen_optimizer = None
        self.p_gen_optimizer = None
        self.m_disc_optimizer = None
        self.p_disc_optimizer = None
        self.gen_loss_fn = None
        self.disc_loss_fn = None
        self.cycle_loss_fn = None
        self.identity_loss_fn = None

        self.tft_layer = None

    def compile(
            self,
            m_gen_optimizer=None,
            p_gen_optimizer=None,
            m_disc_optimizer=None,
            p_disc_optimizer=None,
            gen_loss_fn=None,
            disc_loss_fn=None,
            cycle_loss_fn=None,
            identity_loss_fn=None):

        super(CycleGan, self).compile()
        self.m_gen_optimizer = m_gen_optimizer
        self.p_gen_optimizer = p_gen_optimizer
        self.m_disc_optimizer = m_disc_optimizer
        self.p_disc_optimizer = p_disc_optimizer
        self.gen_loss_fn = gen_loss_fn
        self.disc_loss_fn = disc_loss_fn
        self.cycle_loss_fn = cycle_loss_fn
        self.identity_loss_fn = identity_loss_fn

    def call(self, inputs, training=None, mask=None):
        return self.m_gen(inputs, training=training, mask=mask)

    def train_step(self, batch_data):
        real_monet, real_photo = batch_data
        real_monet = real_monet.pop("image_xf")
        real_photo = real_photo.pop("image_xf")

        with tf.GradientTape(persistent=True) as tape:
            # photo to monet back to photo
            fake_monet = self.m_gen(real_photo, training=True)
            cycled_photo = self.p_gen(fake_monet, training=True)

            # monet to photo back to monet
            fake_photo = self.p_gen(real_monet, training=True)
            cycled_monet = self.m_gen(fake_photo, training=True)

            # generating itself
            same_monet = self.m_gen(real_monet, training=True)
            same_photo = self.p_gen(real_photo, training=True)

            # discriminator used to check input real images
            disc_real_monet = self.m_disc(real_monet, training=True)
            disc_real_photo = self.p_disc(real_photo, training=True)

            # discriminator used to check input fake images
            disc_fake_monet = self.m_disc(fake_monet, training=True)
            disc_fake_photo = self.p_disc(fake_photo, training=True)

            # evaluates generator loss
            monet_gen_loss = self.gen_loss_fn(disc_fake_monet)
            photo_gen_loss = self.gen_loss_fn(disc_fake_photo)

            # evaluates total cycle consistency loss
            total_cycle_loss = self.cycle_loss_fn(
                real_monet, cycled_monet, self.lambda_cycle) + \
                self.cycle_loss_fn(
                real_photo, cycled_photo, self.lambda_cycle)

            # evaluates total generator loss
            total_monet_gen_loss = monet_gen_loss + total_cycle_loss + \
                self.identity_loss_fn(real_monet, same_monet,
                                      self.lambda_cycle)

            total_photo_gen_loss = photo_gen_loss + total_cycle_loss + \
                self.identity_loss_fn(real_photo, same_photo,
                                      self.lambda_cycle)

            # evaluates discriminator loss
            monet_disc_loss = self.disc_loss_fn(disc_real_monet,
                                                disc_fake_monet)
            photo_disc_loss = self.disc_loss_fn(disc_real_photo,
                                                disc_fake_photo)

        # Calculate the gradients for generator and discriminator
        monet_generator_gradients = tape.gradient(
            total_monet_gen_loss, self.m_gen.trainable_variables)

        photo_generator_gradients = tape.gradient(
            total_photo_gen_loss, self.p_gen.trainable_variables)

        monet_discriminator_gradients = tape.gradient(
            monet_disc_loss, self.m_disc.trainable_variables)

        photo_discriminator_gradients = tape.gradient(
            photo_disc_loss, self.p_disc.trainable_variables)

        # Apply the gradients to the optimizer
        self.m_gen_optimizer.apply_gradients(zip(
            monet_generator_gradients, self.m_gen.trainable_variables))

        self.p_gen_optimizer.apply_gradients(zip(
            photo_generator_gradients, self.p_gen.trainable_variables))

        self.m_disc_optimizer.apply_gradients(zip(
            monet_discriminator_gradients, self.m_disc.trainable_variables))

        self.p_disc_optimizer.apply_gradients(zip(
            photo_discriminator_gradients, self.p_disc.trainable_variables))

        return {
            "monet_gen_loss": total_monet_gen_loss,
            "photo_gen_loss": total_photo_gen_loss,
            "monet_disc_loss": monet_disc_loss,
            "photo_disc_loss": photo_disc_loss
        }


def downsample(filters, size, apply_instancenorm=True):
    initializer = tf.random_normal_initializer(0., 0.02)
    gamma_init = tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02)

    result = tf.keras.Sequential()
    result.add(tf.keras.layers.Conv2D(filters, size, strides=2, padding='same',
                                      kernel_initializer=initializer,
                                      use_bias=False))

    if apply_instancenorm:
        result.add(
            tfa.layers.InstanceNormalization(gamma_initializer=gamma_init))

    result.add(tf.keras.layers.LeakyReLU())

    return result


def upsample(filters, size, apply_dropout=False):
    initializer = tf.random_normal_initializer(0., 0.02)
    gamma_init = tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02)

    result = tf.keras.Sequential()
    result.add(tf.keras.layers.Conv2DTranspose(filters, size, strides=2,
                                               padding='same',
                                               kernel_initializer=initializer,
                                               use_bias=False))

    result.add(tfa.layers.InstanceNormalization(gamma_initializer=gamma_init))

    if apply_dropout:
        result.add(tf.keras.layers.Dropout(0.5))

    result.add(tf.keras.layers.ReLU())

    return result


def Generator():
    inputs = tf.keras.layers.Input(shape=[256, 256, 3])

    # bs = batch size
    down_stack = [
        downsample(64, 4, apply_instancenorm=False),  # (bs, 128, 128, 64)
        downsample(128, 4),  # (bs, 64, 64, 128)
        downsample(256, 4),  # (bs, 32, 32, 256)
        downsample(512, 4),  # (bs, 16, 16, 512)
        downsample(512, 4),  # (bs, 8, 8, 512)
        downsample(512, 4),  # (bs, 4, 4, 512)
        downsample(512, 4),  # (bs, 2, 2, 512)
        downsample(512, 4),  # (bs, 1, 1, 512)
    ]

    up_stack = [
        upsample(512, 4, apply_dropout=True),  # (bs, 2, 2, 1024)
        upsample(512, 4, apply_dropout=True),  # (bs, 4, 4, 1024)
        upsample(512, 4, apply_dropout=True),  # (bs, 8, 8, 1024)
        upsample(512, 4),  # (bs, 16, 16, 1024)
        upsample(256, 4),  # (bs, 32, 32, 512)
        upsample(128, 4),  # (bs, 64, 64, 256)
        upsample(64, 4),  # (bs, 128, 128, 128)
    ]

    initializer = tf.random_normal_initializer(0., 0.02)
    # (bs, 256, 256, 3)
    last = tf.keras.layers.Conv2DTranspose(3, 4,
                                           strides=2,
                                           padding='same',
                                           kernel_initializer=initializer,
                                           activation='tanh')

    x = inputs

    # Downsampling through the model
    skips = []
    for down in down_stack:
        x = down(x)
        skips.append(x)

    skips = reversed(skips[:-1])

    # Upsampling and establishing the skip connections
    for up, skip in zip(up_stack, skips):
        x = up(x)
        x = tf.keras.layers.Concatenate()([x, skip])

    x = last(x)

    return tf.keras.Model(inputs=inputs, outputs=x)


def Discriminator():
    initializer = tf.random_normal_initializer(0., 0.02)
    gamma_init = tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.02)

    inp = tf.keras.layers.Input(shape=[256, 256, 3], name='input_image')

    x = inp

    down1 = downsample(64, 4, False)(x)  # (bs, 128, 128, 64)
    down2 = downsample(128, 4)(down1)  # (bs, 64, 64, 128)
    down3 = downsample(256, 4)(down2)  # (bs, 32, 32, 256)

    zero_pad1 = tf.keras.layers.ZeroPadding2D()(down3)  # (bs, 34, 34, 256)
    conv = tf.keras.layers.Conv2D(512, 4, strides=1,
                                  kernel_initializer=initializer,
                                  use_bias=False)(
        zero_pad1)  # (bs, 31, 31, 512)

    norm1 = tfa.layers.InstanceNormalization(gamma_initializer=gamma_init)(
        conv)

    leaky_relu = tf.keras.layers.LeakyReLU()(norm1)

    zero_pad2 = tf.keras.layers.ZeroPadding2D()(
        leaky_relu)  # (bs, 33, 33, 512)

    last = tf.keras.layers.Conv2D(1, 4, strides=1,
                                  kernel_initializer=initializer)(
        zero_pad2)  # (bs, 30, 30, 1)

    return tf.keras.Model(inputs=inp, outputs=last)


def discriminator_loss(real, generated):
    real_loss = tf.keras.losses.BinaryCrossentropy(
        from_logits=True, reduction=tf.keras.losses.Reduction.NONE)(
        tf.ones_like(real), real)

    generated_loss = tf.keras.losses.BinaryCrossentropy(
        from_logits=True, reduction=tf.keras.losses.Reduction.NONE)(
        tf.zeros_like(generated), generated)

    total_disc_loss = real_loss + generated_loss

    return total_disc_loss * 0.5


def generator_loss(generated):
    return tf.keras.losses.BinaryCrossentropy(
        from_logits=True, reduction=tf.keras.losses.Reduction.NONE)(
        tf.ones_like(generated), generated)


def calc_cycle_loss(real_image, cycled_image, lamb):
    loss1 = tf.reduce_mean(tf.abs(real_image - cycled_image))
    return lamb * loss1


def identity_loss(real_image, same_image, lamb):
    loss = tf.reduce_mean(tf.abs(real_image - same_image))
    return lamb * 0.5 * loss


class TensorBoardImage(tf.keras.callbacks.Callback):
    def __init__(self, test_data: tf.data.Dataset, log_dir):
        super().__init__()
        self.log_dir = log_dir
        self.test_data = test_data
        self.writer = tf.summary.create_file_writer(self.log_dir)

    def on_epoch_end(self, epoch, logs=None):
        # Get image by transforming dataset to numpy iterator,
        # requires eager execution mode
        data = self.test_data.take(1).as_numpy_iterator()

        for element in data:
            output = self.model(element)
            normal_output = (output + 1) * 127.5
            with self.writer.as_default():
                tf.summary.image("GAN Example image",
                                 tf.cast(normal_output, tf.uint8),
                                 step=epoch)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Dict

import tensorflow as tf

from zenml.steps.preprocesser import BasePreprocesserStep
from zenml.utils import naming_utils


def decode_and_reshape_image(input_):
    densed = tf.sparse.to_dense(input_)
    image = tf.map_fn(lambda x: tf.io.decode_image(x[0], channels=3), densed,
                      dtype=tf.uint8)
    image = (tf.cast(image, tf.float32) / 127.5) - 1
    image = tf.reshape(image, [-1, 256, 256, 3])
    return image


class GANPreprocessor(BasePreprocesserStep):

    def __init__(self, **unused_kwargs):

        super(GANPreprocessor, self).__init__(**unused_kwargs)

    def preprocessing_fn(self, inputs: Dict):

        outputs = {}

        for k, v in inputs.items():
            if k == "image":
                result = decode_and_reshape_image(v)
                result = tf.cast(result, dtype=tf.float32)
                outputs[naming_utils.transformed_feature_name(k)] = result

        return outputs



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CLI for pipelines."""

from typing import Text

import click
from tabulate import tabulate

from zenml.cli.cli import cli
from zenml.cli.utils import error, pretty_print, pass_repo
from zenml.pipelines import TrainingPipeline
from zenml.repo import Repository
from zenml.utils.yaml_utils import read_yaml


@cli.group()
def pipeline():
    """Pipeline group"""
    pass


@pipeline.command('compare')
@pass_repo
def compare_training_runs(repo: Repository):
    """Compares pipelines in repo"""
    click.echo('Comparing training pipelines in repo: Starting app..')
    repo.compare_training_runs()


@pipeline.command('list')
@pass_repo
def list_pipelines(repo: Repository):
    """Lists pipelines in the current repository."""
    try:
        pipelines = repo.get_pipelines()

        names = [p.name for p in pipelines]
        types = [p.PIPELINE_TYPE for p in pipelines]
        statuses = [p.get_status() for p in pipelines]
        cache_enabled = [p.enable_cache for p in pipelines]
        filenames = [p.file_name for p in pipelines]

        headers = ["name", "type", "cache enabled", "status", "file name"]

        click.echo(tabulate(zip(names, types, cache_enabled,
                                statuses, filenames),
                            headers=headers))
    except Exception as e:
        error(e)


@pipeline.command('get')
@click.argument('pipeline_name')
@pass_repo
def get_pipeline_by_name(repo: Repository, pipeline_name: Text):
    """
    Gets pipeline from current repository by matching a name against a
    pipeline name in the repository.
    """
    try:
        p = repo.get_pipeline_by_name(pipeline_name)
    except Exception as e:
        error(e)
        return

    pretty_print(p)


@pipeline.command('run')
@click.argument('path_to_config')
def run_pipeline(path_to_config: Text):
    """
    Runs pipeline specified by the given config YAML object.

    Args:
        path_to_config: Path to config of the designated pipeline.
         Has to be matching the YAML file name.
    """
    # config has metadata store, backends and artifact store,
    # so no need to specify them
    print(path_to_config)
    try:
        config = read_yaml(path_to_config)
        p: TrainingPipeline = TrainingPipeline.from_config(config)
        p.run()
    except Exception as e:
        error(e)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""File utilities"""

import fnmatch
import os
import tarfile
from pathlib import Path
from typing import Text, Callable

from tensorflow.python.lib.io import file_io
from tfx.utils.io_utils import _REMOTE_FS_PREFIX, load_csv_column_names


# TODO: [TFX] [LOW] Unnecessary dependency here


def walk(dir_path):
    """Walks down the dir_path"""
    return file_io.walk_v2(dir_path)


def is_root(path: Text):
    """
    Returns true if path has no parent in local filesystem.

    Args:
        path (str): Local path in filesystem.
    """
    return Path(path).parent == Path(path)


def is_dir(dir_path: Text):
    """
    Returns true if dir_path points to a dir.

    Args:
        dir_path (str): Local path in filesystem.
    """
    return file_io.is_directory_v2(dir_path)


def find_files(dir_path, pattern):
    """
    Find files in a directory that match pattern.

    Args:
        dir_path: Path to directory.
        pattern: pattern like *.png.
    """
    for root, dirs, files in walk(dir_path):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename


def is_remote(path: Text):
    """
    Returns True if path exists remotely.

    Args:
        path (str): Any path.
    """
    return any([path.startswith(prefix) for prefix in
                _REMOTE_FS_PREFIX])


def is_gcs_path(path: Text):
    """
    Returns True if path is on Google Cloud Storage.

    Args:
        path (str): Any path.
    """
    return path.startswith('gs://')


def list_dir(dir_path: Text, only_file_names: bool = False):
    """
    Returns a list of files under dir.

    Args:
        dir_path (str): Path in filesystem.
        only_file_names (bool): Returns only file names if True.
    """
    return [os.path.join(dir_path, f) if not only_file_names else f
            for f in file_io.list_directory_v2(dir_path)]


def create_file_if_not_exists(file_path: Text, file_contents: Text):
    """
    Creates directory if it does not exist.

    Args:
        file_path (str): Local path in filesystem.
        file_contents (str): Contents of file.
    """
    if not file_io.file_exists_v2(file_path):
        file_io.write_string_to_file(file_path, file_contents)


def append_file(file_path: Text, file_contents: Text):
    """
    Appends file_contents to file.

    Args:
        file_path (str): Local path in filesystem.
        file_contents (str): Contents of file.
    """
    with file_io.FileIO(file_path, mode='a') as f:
        f.write(file_contents)


def create_dir_if_not_exists(dir_path: Text):
    """
    Creates directory if it does not exist.

    Args:
        dir_path (str): Local path in filesystem.
    """
    if not file_io.is_directory_v2(dir_path):
        file_io.create_dir_v2(dir_path)


def create_dir_recursive_if_not_exists(dir_path: Text):
    """
    Creates directory recursively if it does not exist.

    Args:
        dir_path (str): Local path in filesystem.
    """
    if not file_io.is_directory_v2(dir_path):
        file_io.recursive_create_dir_v2(dir_path)


def resolve_relative_path(path: Text):
    """
    Takes relative path and resolves it absolutely.

    Args:
        path (str): Local path in filesystem.
    """
    if is_remote(path):
        return path
    return str(Path(path).resolve())


def file_exists(path: Text):
    """
    Returns true if file exists at path.

    Args:
        path (str): Local path in filesystem.
    """
    return file_io.file_exists_v2(path)


def copy(source: Text, destination: Text, overwrite: bool = False):
    """
    Copies dir from source to destination.

    Args:
        source (str): Path to copy from.
        destination (str): Path to copy to.
        overwrite: boolean, if false, then throws an error before overwrite.
    """
    return file_io.copy_v2(source, destination, overwrite)


def copy_dir(source_dir: Text, destination_dir: Text, overwrite: bool = False):
    """
    Copies dir from source to destination.

    Args:
        source_dir (str): Path to copy from.
        destination_dir (str): Path to copy to.
        overwrite: boolean, if false, then throws an error before overwrite.
    """
    for f in list_dir(source_dir):
        p = Path(f)
        destination_name = os.path.join(destination_dir, p.name)
        if is_dir(f):
            copy_dir(f, destination_name, overwrite)
        else:
            create_dir_recursive_if_not_exists(
                str(Path(destination_name).parent))
            copy(f, destination_name, overwrite)


def move(source: Text, destination: Text, overwrite: bool = False):
    """
    Moves dir from source to destination. Can be used to rename.

    Args:
        source (str): Local path to copy from.
        destination (str): Local path to copy to.
        overwrite: boolean, if false, then throws an error before overwrite.
    """
    return file_io.rename_v2(source, destination, overwrite)


def rm_dir(dir_path: Text):
    """
    Deletes dir recursively. Dangerous operation.

    Args:
        dir_path (str): Dir to delete.
    """
    file_io.delete_recursively_v2(dir_path)


def rm_file(file_path: Text):
    """
    Deletes file. Dangerous operation.

    Args:
        file_path (str): Path of file to delete.
    """
    if not file_exists(file_path):
        raise Exception(f'{file_path} does not exist!')
    return file_io.delete_file_v2(file_path)


def read_file_contents(file_path: Text):
    """
    Reads contents of file.

    Args:
        file_path (str): Path to file.
    """
    if not file_exists(file_path):
        raise Exception(f'{file_path} does not exist!')
    return file_io.read_file_to_string(file_path)


def write_file_contents(file_path: Text, content: Text):
    """
    Writes contents of file.

    Args:
        file_path (str): Path to file.
        content (str): Contents of file.
    """
    return file_io.write_string_to_file(file_path, content)


def get_grandparent(dir_path: Text):
    """
    Get grandparent of dir.

    Args:
        dir_path (str): Path to directory.
    """
    return Path(dir_path).parent.stem


def get_parent(dir_path: Text):
    """
    Get parent of dir.

    Args:
        dir_path (str): Path to directory.
    """
    return Path(dir_path).stem


def load_csv_header(csv_path: Text):
    """
    Gets header column of csv and returns list.

    Args:
        csv_path (str): Path to csv file.
    """
    return load_csv_column_names(csv_path)


def create_tarfile(source_dir: Text, output_filename: Text = 'zipped.tar.gz',
                   exclude_function: Callable = None):
    """
    Create a compressed representation of source_dir.

    Args:
        source_dir: path to source dir
        output_filename: name of outputted gz
        exclude_function: function that determines whether to exclude file
    """
    if exclude_function is None:
        # default is to exclude the .zenml directory
        def exclude_function(tarinfo):
            filename = tarinfo.name
            if '.zenml/' in filename:
                return None
            elif 'venv/' in filename:
                return None
            else:
                return tarinfo

    with tarfile.open(output_filename, "w:gz") as tar:
        tar.add(source_dir, arcname='', filter=exclude_function)


def extract_tarfile(source_tar: Text, output_dir: Text):
    """
    Untars a compressed tar file to output_dir.

    Args:
        source_tar: path to a tar compressed file
        output_dir: directory where to uncompress
    """
    if is_remote(source_tar):
        raise NotImplementedError('Use local tars for now.')

    with tarfile.open(source_tar, "r:gz") as tar:
        tar.extractall(output_dir)



#  Copyright (c) maiot GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.

from typing import Text, Dict, Any, Optional

from tfx.dsl.components.base.base_component import BaseComponent
from tfx.dsl.components.base.executor_spec import ExecutorClassSpec
from tfx.types import standard_artifacts, Channel
from tfx.types.component_spec import ComponentSpec, ExecutionParameter, \
    ChannelParameter

from zenml.components.bulk_inferrer.constants import MODEL, EXAMPLES, \
    MODEL_BLESSING, PREDICTIONS
from zenml.components.bulk_inferrer.executor import BulkInferrerExecutor
from zenml.standards.standard_keys import StepKeys


class BulkInferrerSpec(ComponentSpec):
    PARAMETERS = {
        StepKeys.SOURCE: ExecutionParameter(type=Text),
        StepKeys.ARGS: ExecutionParameter(type=Dict[Text, Any]),
    }
    INPUTS = {
        MODEL: ChannelParameter(type=standard_artifacts.Model, optional=True),
        EXAMPLES: ChannelParameter(type=standard_artifacts.Examples),
        MODEL_BLESSING: ChannelParameter(type=standard_artifacts.ModelBlessing,
                                         optional=True)

    }
    OUTPUTS = {
        PREDICTIONS: ChannelParameter(type=standard_artifacts.Examples)
    }


class BulkInferrer(BaseComponent):
    SPEC_CLASS = BulkInferrerSpec
    EXECUTOR_SPEC = ExecutorClassSpec(BulkInferrerExecutor)

    def __init__(self,
                 source: Text,
                 source_args: Dict[Text, Any],
                 model: Optional[ChannelParameter] = None,
                 model_blessing: Optional[ChannelParameter] = None,
                 instance_name: Optional[Text] = None,
                 examples: Optional[ChannelParameter] = None,
                 predictions: Optional[ChannelParameter] = None):
        """
        Interface for all DataGen components, the main component responsible
        for reading data and converting to TFRecords. This is how we handle
        versioning data for now.
        Args:
            source:
            source_args:
            model:
            model_blessing:
            instance_name:
            examples:
            predictions:
        """
        examples = examples or Channel(type=standard_artifacts.Examples)
        predictions = predictions or Channel(type=standard_artifacts.Examples)

        # Initiate the spec and create instance
        spec = self.SPEC_CLASS(source=source,
                               args=source_args,
                               model=model,
                               model_blessing=model_blessing,
                               examples=examples,
                               predictions=predictions)

        super(BulkInferrer, self).__init__(spec=spec,
                                           instance_name=instance_name)



#  Copyright (c) maiot GmbH 2020. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Some split step utilities are implemented here."""

from typing import Text, List, Dict, Union

import tensorflow as tf

CategoricalValue = Union[Text, int]


def get_categorical_value(example: tf.train.Example, cat_col: Text):
    """
    Helper function to get the categorical value from a tf.train.Example.

    Args:
        example: tf.train.Example, data point in proto format.
        cat_col: Name of the categorical feature of which to extract the
        value from.

    Returns:
        value: The categorical value found in the `cat_col` feature inside
        the tf.train.Example.

    Raises:
        AssertionError: If the `cat_col` feature is not present in the
        tf.train.Example.
    """

    cat_feature = example.features.feature[cat_col]

    possible_types = ["bytes", "float", "int64"]

    for datatype in possible_types:
        value_list = getattr(cat_feature, datatype + "_list")

        if value_list.value:
            value = value_list.value[0]
            if hasattr(value, "decode"):
                return value.decode()
            else:
                return value

    raise AssertionError(f"Error: Feature {cat_col} does not exist in "
                         f"this data.")


def partition_cat_list(cat_list: List[CategoricalValue],
                       c_ratio: Dict[Text, float]):
    """
    Helper to split a category list by the entries in a category split dict.

    Args:
        cat_list: List of categorical values found in the categorical column.
        c_ratio: Dict {fold: ratio} mapping the ratio of all categories to
         split folds.

    Returns:
        cat_dict: Dict {fold: categorical_list} mapping lists of categorical
         values in the data to their designated split folds.
    """

    cat_dict = {}

    num_cats = len(cat_list)
    sum_ratios = sum(c_ratio.values())
    ratio = 0

    # This might produce unexpected results if the number of categories
    # is lower than the number of folds.
    for fold, fold_pct in c_ratio.items():
        left_bound = round(num_cats * ratio)
        ratio += fold_pct / sum_ratios
        right_bound = round(num_cats * ratio)
        # write categories for fold to the cat_list dict
        cat_dict[fold] = cat_list[left_bound:right_bound]

    return cat_dict



# ZenML constants
TEST_RESULTS = 'test_results'

# TFX constants
RUN_FN = 'run_fn'
CUSTOM_CONFIG = 'custom_config'
EXAMPLES = 'examples'
SCHEMA = 'schema'
BASE_MODEL = 'base_model'
TRANSFORM_GRAPH = 'transform_graph'
HYPERPARAMETERS = 'hyperparameters'
MODEL = 'model'
MODEL_RUN = 'model_run'



import base64
import json
import os
from typing import Dict, Text

import boto3

from zenml.constants import ZENML_BASE_IMAGE_NAME, AWS_ENTRYPOINT

AWS_ACCESS_KEY_ID = 'AWS_ACCESS_KEY_ID'
AWS_SECRET_ACCESS_KEY = 'AWS_SECRET_ACCESS_KEY'
AWS_REGION = 'AWS_REGION'

SHARED_CREDENTIALS_FILE = 'shared-credentials-file'


def _get_container_params(config):
    config_encoded = base64.b64encode(json.dumps(config).encode())
    return f'python -m {AWS_ENTRYPOINT} run_pipeline --config_b64 ' \
           f'{config_encoded.decode()}'


def get_startup_script(config: Dict,
                       region: Text,
                       zenml_image: Text = ZENML_BASE_IMAGE_NAME):
    c_params = _get_container_params(config)
    return f"#!/bin/bash\n" \
           f"mkdir -p /tmp/aws_config\n" \
           f"touch /tmp/aws_config/config\n" \
           f"echo \"[default]\">>/tmp/aws_config/config\n" \
           f"echo \"region = {region}\">>/tmp/aws_config/config\n" \
           f"sudo HOME=/home/root docker run --net=host " \
           f"--env AWS_REGION={region} -v /tmp/aws_config:/root/.aws " \
           f"{zenml_image} {c_params}\n" \
           f"instanceId=$(curl http://169.254.169.254/latest/meta-data/instance-id/)\n" \
           f"/usr/bin/aws ec2 terminate-instances --instance-ids $instanceId --region {region}"


def setup_session():
    session = boto3.Session()
    credentials = session.get_credentials()
    if credentials.method == 'shared-credentials-file':
        os.environ[AWS_ACCESS_KEY_ID] = credentials.access_key
        os.environ[AWS_SECRET_ACCESS_KEY] = credentials.secret_key
    return session


def setup_region(region: Text = None):
    if region is None:
        if AWS_REGION in os.environ:
            pass
        else:
            session = boto3.Session()
            if session.region_name is None:
                os.environ[AWS_REGION] = 'eu-central-1'
            else:
                os.environ[AWS_REGION] = session.region_name
    else:
        os.environ[AWS_REGION] = region
    return os.environ[AWS_REGION]
